#!/usr/bin/env python3
# Import needed libraries
import numpy as np
import openmdao.api as om

from wisdem.commonse.fileIO import save_data
from wisdem.drivetrainse.drivetrain import DrivetrainSE

opt_flag = True
# ---

# Set input options
opt = {}
opt["WISDEM"] = {}
opt["WISDEM"]["n_dlc"] = 1
opt["WISDEM"]["DriveSE"] = {}
opt["WISDEM"]["DriveSE"]["direct"] = False
opt["WISDEM"]["DriveSE"]["use_gb_torque_density"] = False
opt["WISDEM"]["DriveSE"]["hub"] = {}
opt["WISDEM"]["DriveSE"]["hub"]["hub_gamma"] = 2.0
opt["WISDEM"]["DriveSE"]["hub"]["spinner_gamma"] = 1.5
opt["WISDEM"]["DriveSE"]["gamma_f"] = 1.35
opt["WISDEM"]["DriveSE"]["gamma_m"] = 1.3
opt["WISDEM"]["DriveSE"]["gamma_n"] = 1.0
opt["WISDEM"]["RotorSE"] = {}
opt["WISDEM"]["RotorSE"]["n_pc"] = 20
opt["materials"] = {}
opt["materials"]["n_mat"] = 4
opt["flags"] = {}
opt["flags"]["generator"] = False
# ---

# Initialize OpenMDAO problem
prob = om.Problem(reports=False)
prob.model = DrivetrainSE(modeling_options=opt)
# ---

# If performing optimization, set up the optimizer and problem formulation
if opt_flag:
    # Choose the optimizer to use
    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options["optimizer"] = "SLSQP"
    prob.driver.options["tol"] = 1e-2
    prob.driver.options["maxiter"] = 5

    # Add objective
    prob.model.add_objective("nacelle_mass", scaler=1e-6)

    # Add design variables, in this case the tower diameter and wall thicknesses
    prob.model.add_design_var("L_12", lower=0.1, upper=5.0)
    prob.model.add_design_var("L_h1", lower=0.1, upper=5.0)
    prob.model.add_design_var("L_hss", lower=0.1, upper=5.0)
    prob.model.add_design_var("hub_diameter", lower=2.0, upper=5.0)
    prob.model.add_design_var("lss_diameter", lower=0.5, upper=6.0)
    prob.model.add_design_var("lss_wall_thickness", lower=4e-3, upper=5e-1, ref=1e-2)
    prob.model.add_design_var("hss_diameter", lower=0.5, upper=6.0)
    prob.model.add_design_var("hss_wall_thickness", lower=4e-3, upper=5e-1, ref=1e-2)
    prob.model.add_design_var("bedplate_web_thickness", lower=4e-3, upper=5e-1, ref=1e-2)
    prob.model.add_design_var("bedplate_flange_thickness", lower=4e-3, upper=5e-1, ref=1e-2)
    prob.model.add_design_var("bedplate_flange_width", lower=0.1, upper=2.0)

    # Add constraints on the tower design
    prob.model.add_constraint("constr_lss_vonmises", upper=1.0)
    prob.model.add_constraint("constr_hss_vonmises", upper=1.0)
    prob.model.add_constraint("constr_bedplate_vonmises", upper=1.0)
    prob.model.add_constraint("constr_mb1_defl", upper=1.0)
    prob.model.add_constraint("constr_mb2_defl", upper=1.0)
    prob.model.add_constraint("constr_shaft_deflection", upper=1.0)
    prob.model.add_constraint("constr_shaft_angle", upper=1.0)
    prob.model.add_constraint("constr_stator_deflection", upper=1.0)
    prob.model.add_constraint("constr_stator_angle", upper=1.0)
    prob.model.add_constraint("constr_hub_diameter", lower=0.0)
    prob.model.add_constraint("constr_length", lower=0.0)
    prob.model.add_constraint("constr_height", lower=0.0)
    # ---


# Set up the OpenMDAO problem
prob.setup()
# ---

# Set input values
prob.set_val("machine_rating", 5.0, units="MW")
prob["upwind"] = True
prob["n_blades"] = 3
prob["rotor_diameter"] = 126.0
prob["D_top"] = 3.87
prob["minimum_rpm"] = 6.9
prob["rated_rpm"] = 12.1

# Loading from rotor
prob["F_hub"] = np.array([1125044.07614847, -7098.0872533, -7022.79756034]).reshape((3, 1))
prob["M_hub"] = np.array([10515165.10636333, 945938.60268626, 1042828.16100417]).reshape((3, 1))
# ---

# Blade properties and hub design options
prob["blade_mass"] = 16403.0
prob["blades_mass"] = 3 * prob["blade_mass"]
prob["pitch_system.BRFM"] = 14239550.0
prob["pitch_system_scaling_factor"] = 0.54
prob["blade_root_diameter"] = 3.542
prob["flange_t2shell_t"] = 4.0
prob["flange_OD2hub_D"] = 0.5
prob["flange_ID2flange_OD"] = 0.8
prob["hub_in2out_circ"] = 1.2
prob["hub_stress_concentration"] = 2.5
prob["n_front_brackets"] = 3
prob["n_rear_brackets"] = 3
prob["clearance_hub_spinner"] = 1.0
prob["spin_hole_incr"] = 1.2
prob["spinner_gust_ws"] = 70.0
prob["hub_diameter"] = 3.0
prob["blades_I"] = np.r_[36494351.0, 17549243.0, 14423664.0, np.zeros(3)]
# ---

# Drivetrain configuration and sizing inputs
prob["bear1.bearing_type"] = "CARB"
prob["bear2.bearing_type"] = "SRB"
prob["L_12"] = 0.368
prob["L_h1"] = 1.912
prob["L_hss"] = 1.5
prob["L_generator"] = 2.0
prob["L_gearbox"] = 1.5
prob["overhang"] = 5.0
prob["drive_height"] = 2.3
prob["tilt"] = 5.0

prob["planet_numbers"] = np.array([3, 3, 0])
prob["gear_configuration"] = "eep"
prob["gear_ratio"] = 96.0

myones = np.ones(2)
prob["lss_diameter"] = 1.0 * myones
prob["hss_diameter"] = 0.5 * myones
prob["lss_wall_thickness"] = 0.288 * myones
prob["hss_wall_thickness"] = 0.1 * myones
prob["bedplate_web_thickness"] = 0.1
prob["bedplate_flange_thickness"] = 0.1
prob["bedplate_flange_width"] = 1.0
prob["bear1.D_shaft"] = 2.2
prob["bear2.D_shaft"] = 2.2
prob["shaft_deflection_allowable"] = 1e-4
prob["shaft_angle_allowable"] = 1e-3
prob["stator_deflection_allowable"] = 1e-4
prob["stator_angle_allowable"] = 1e-3
# ---

# Material properties
prob["E_mat"] = np.c_[200e9 * np.ones(3), 205e9 * np.ones(3), 118e9 * np.ones(3), [4.46e10, 1.7e10, 1.67e10]].T
prob["G_mat"] = np.c_[79.3e9 * np.ones(3), 80e9 * np.ones(3), 47.6e9 * np.ones(3), [3.27e9, 3.48e9, 3.5e9]].T
prob["Xt_mat"] = np.c_[450e6 * np.ones(3), 814e6 * np.ones(3), 310e6 * np.ones(3), [6.092e8, 3.81e7, 1.529e7]].T
prob["rho_mat"] = np.r_[7800.0, 7850.0, 7200.0, 1940.0]
prob["Xy_mat"] = np.r_[345e6, 485e6, 265e6, 18.9e6]
prob["wohler_exp_mat"] = 1e1 * np.ones(4)
prob["wohler_A_mat"] = 1e1 * np.ones(4)
prob["unit_cost_mat"] = np.r_[0.7, 0.9, 0.5, 1.9]
prob["lss_material"] = prob["hss_material"] = "steel_drive"
prob["bedplate_material"] = "steel"
prob["hub_material"] = "cast_iron"
prob["spinner_material"] = "glass_uni"
prob["material_names"] = ["steel", "steel_drive", "cast_iron", "glass_uni"]
# ---

# Run the analysis or optimization
if opt_flag:
    prob.model.approx_totals()
    prob.run_driver()
else:
    prob.run_model()
save_data("drivetrain_example", prob)
# ---

# Display all inputs and outputs
# prob.model.list_inputs(units=True)
# prob.model.list_outputs(units=True)

# Print out the objective, design variables and constraints
print("nacelle_mass:", prob["nacelle_mass"])
print("")
print("L_h1:", prob["L_h1"])
print("L_12:", prob["L_12"])
print("L_lss:", prob["L_lss"])
print("L_hss:", prob["L_hss"])
print("L_generator:", prob["L_generator"])
print("L_gearbox:", prob["L_gearbox"])
print("L_bedplate:", prob["L_bedplate"])
print("H_bedplate:", prob["H_bedplate"])
print("hub_diameter:", prob["hub_diameter"])
print("lss_diameter:", prob["lss_diameter"])
print("lss_wall_thickness:", prob["lss_wall_thickness"])
print("hss_diameter:", prob["hss_diameter"])
print("hss_wall_thickness:", prob["hss_wall_thickness"])
print("bedplate_web_thickness:", prob["bedplate_web_thickness"])
print("bedplate_flange_thickness:", prob["bedplate_flange_thickness"])
print("bedplate_flange_width:", prob["bedplate_flange_width"])
print("")
print("constr_lss_vonmises:", prob["constr_lss_vonmises"].flatten())
print("constr_hss_vonmises:", prob["constr_hss_vonmises"].flatten())
print("constr_bedplate_vonmises:", prob["constr_bedplate_vonmises"].flatten())
print("constr_mb1_defl:", prob["constr_mb1_defl"])
print("constr_mb2_defl:", prob["constr_mb2_defl"])
print("constr_shaft_deflection:", prob["constr_shaft_deflection"])
print("constr_shaft_angle:", prob["constr_shaft_angle"])
print("constr_stator_deflection:", prob["constr_stator_deflection"])
print("constr_stator_angle:", prob["constr_stator_angle"])
print("constr_hub_diameter:", prob["constr_hub_diameter"])
print("constr_length:", prob["constr_length"])
print("constr_height:", prob["constr_height"])
# ---



#!/usr/bin/env python3

# Import the libraries
import openmdao.api as om

from wisdem.plant_financese.plant_finance import PlantFinance

# Initialize the OpenMDAO instance
mygroup = om.Group()
mygroup.add_subsystem("main", PlantFinance(), promotes=["*"])

prob = om.Problem(reports=False)
prob.model = mygroup
prob.setup()

# Set variable inputs with intended units
prob.set_val("machine_rating", 2.32e3, units="kW")
prob.set_val("tcc_per_kW", 1093, units="USD/kW")
prob.set_val("turbine_number", 87)
prob.set_val("opex_per_kW", 43.56, units="USD/kW/yr")
prob.set_val("fixed_charge_rate", 0.079216644)
prob.set_val("bos_per_kW", 517.0, units="USD/kW")
prob.set_val("wake_loss_factor", 0.15)
prob.set_val("turbine_aep", 9915.95e3, units="kW*h")

# Run the model once
prob.run_model()

# Print all inputs and outputs to the screen
prob.model.list_inputs(units=True)
prob.model.list_outputs(units=True)



import os
import sys
import time

from wisdem.commonse.mpi_tools import MPI
from wisdem.glue_code.runWISDEM import run_wisdem

## File management
run_dir = os.path.dirname(os.path.realpath(__file__)) + os.sep
wisdem_examples = os.path.dirname(os.path.dirname(run_dir))
fname_wt_input_oc3 = os.path.join(wisdem_examples, "09_floating", "nrel5mw-spar_oc3.yaml")
fname_modeling_options = os.path.join(wisdem_examples, "09_floating", "modeling_options_noRNA.yaml")
fname_analysis_options = run_dir + os.sep + "analysis_options_spar.yaml"


wt_opt, modeling_options, opt_options = run_wisdem(fname_wt_input_oc3, fname_modeling_options, fname_analysis_options)



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

import os

import pandas as pd

DIR = os.path.split(__file__)[0]
_fp = os.path.join(DIR, "test_weather.csv")
test_weather = pd.read_csv(_fp, parse_dates=["datetime"]).set_index("datetime").to_records()



""" Unit tests for the NLOpt Driver."""

import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.general_utils import run_driver
from openmdao.test_suite.components.sellar import SellarDerivatives, SellarDerivativesGrouped
from openmdao.test_suite.groups.sin_fitter import SineFitter
from openmdao.test_suite.components.paraboloid import Paraboloid
from openmdao.test_suite.components.simple_comps import NonSquareArrayComp
from openmdao.test_suite.components.expl_comp_array import TestExplCompArrayDense as examp

from wisdem.optimization_drivers.nlopt_driver import NLoptDriver

try:
    import nlopt
except ImportError:
    nlopt = None


def rastrigin(x):
    a = 10  # constant
    return np.sum(np.square(x) - a * np.cos(2 * np.pi * x)) + a * np.size(x)


@unittest.skipIf(nlopt is None, "only run if NLopt is installed.")
class TestNLoptDriver(unittest.TestCase):
    def test_driver_supports(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver(optimizer="LD_SLSQP", tol=1e-9)

        with self.assertRaises(KeyError) as raises_msg:
            prob.driver.supports["equality_constraints"] = False

        exception = raises_msg.exception

        msg = "NLoptDriver: Tried to set read-only option 'equality_constraints'."

        self.assertEqual(exception.args[0], msg)

    def test_compute_totals_basic_return_array(self):
        # Make sure 'array' return_format works.

        prob = om.Problem(reports=False)
        model = prob.model
        model.add_subsystem("p1", om.IndepVarComp("x", 0.0), promotes=["x"])
        model.add_subsystem("p2", om.IndepVarComp("y", 0.0), promotes=["y"])
        model.add_subsystem("comp", Paraboloid(), promotes=["x", "y", "f_xy"])

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup(check=False, mode="fwd")
        prob.set_solver_print(level=0)

        failed = prob.run_driver()

        of = ["f_xy"]
        wrt = ["x", "y"]
        derivs = prob.compute_totals(of=of, wrt=wrt, return_format="array")

        assert_near_equal(derivs[0, 0], -6.0, 1e-6)
        assert_near_equal(derivs[0, 1], 8.0, 1e-6)

        prob.setup(check=False, mode="rev")

        prob.run_model()

        of = ["f_xy"]
        wrt = ["x", "y"]
        derivs = prob.compute_totals(of=of, wrt=wrt, return_format="array")

        assert_near_equal(derivs[0, 0], -6.0, 1e-6)
        assert_near_equal(derivs[0, 1], 8.0, 1e-6)

    def test_compute_totals_return_array_non_square(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("px", om.IndepVarComp(name="x", val=np.ones((2,))))
        comp = model.add_subsystem("comp", NonSquareArrayComp())
        model.connect("px.x", "comp.x1")

        model.add_design_var("px.x")
        model.add_objective("px.x")
        model.add_constraint("comp.y1")
        model.add_constraint("comp.y2")

        prob.setup(check=False, mode="auto")

        failed = prob.run_driver()

        derivs = prob.compute_totals(of=["comp.y1"], wrt=["px.x"], return_format="array")

        J = comp.JJ[0:3, 0:2]
        assert_near_equal(J, derivs, 1.0e-3)

        # Support for a name to be in 'of' and 'wrt'

        derivs = prob.compute_totals(of=["comp.y2", "px.x", "comp.y1"], wrt=["px.x"], return_format="array")

        assert_near_equal(J, derivs[3:, :], 1.0e-3)
        assert_near_equal(comp.JJ[3:4, 0:2], derivs[0:1, :], 1.0e-3)
        assert_near_equal(np.eye(2), derivs[1:3, :], 1.0e-3)

    def test_deriv_wrt_self(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("px", om.IndepVarComp(name="x", val=np.ones((2,))))

        model.add_design_var("px.x")
        model.add_objective("px.x")

        prob.setup()

        failed = prob.run_driver()

        # Support for a name to be in 'of' and 'wrt'

        J = prob.driver._compute_totals(of=["px.x"], wrt=["px.x"], return_format="array")

        assert_near_equal(J, np.eye(2), 1.0e-3)

    def test_optimizer_simple_paraboloid_unconstrained(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver(optimizer="LD_SLSQP", tol=1e-9)

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        failed = prob.run_driver()

        assert_near_equal(prob["x"], 6.66666667, 1e-6)
        assert_near_equal(prob["y"], -7.3333333, 1e-6)

    def test_simple_paraboloid_unconstrained(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        failed = prob.run_driver()

        assert_near_equal(prob["x"], 6.66666667, 1e-6)
        assert_near_equal(prob["y"], -7.3333333, 1e-6)

    def test_simple_paraboloid_unconstrained_LN_COBYLA(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LN_COBYLA"
        prob.driver.options["tol"] = 1e-12

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        failed = prob.run_driver()

        assert_near_equal(prob["x"], 6.66666667, 1e-6)
        assert_near_equal(prob["y"], -7.3333333, 1e-6)

    def test_simple_paraboloid_upper(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", upper=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        assert_near_equal(prob["x"], 7.16667, 1e-6)
        assert_near_equal(prob["y"], -7.833334, 1e-6)

    def test_simple_paraboloid_lower(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)

        model.add_objective("f_xy")
        model.add_constraint("c", lower=15.0)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        assert_near_equal(prob["x"], 7.16667, 1e-6)
        assert_near_equal(prob["y"], -7.833334, 1e-6)

    def test_simple_paraboloid_equality(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        assert_near_equal(prob["x"], 7.16667, 1e-4)
        assert_near_equal(prob["y"], -7.833334, 1e-4)

    def test_missing_objective(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("x", om.IndepVarComp("x", 2.0), promotes=["*"])
        model.add_subsystem("f_x", Paraboloid(), promotes=["*"])

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"

        prob.model.add_design_var("x", lower=0)

        prob.setup()

        with self.assertRaises(Exception) as raises_msg:
            prob.run_driver()

        exception = raises_msg.exception

        msg = "Driver requires objective to be declared"

        self.assertEqual(exception.args[0], msg)

    def test_simple_paraboloid_double_sided_low(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", lower=-11.0, upper=-10.0)

        prob.setup()

        failed = prob.run_driver()

        assert_near_equal(prob["y"] - prob["x"], -11.0, 1e-6)

    def test_simple_paraboloid_double_sided_high(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", lower=10.0, upper=11.0)

        prob.setup()

        failed = prob.run_driver()

        assert_near_equal(prob["x"] - prob["y"], 11.0, 1e-6)

    def test_simple_array_comp2D(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("widths", np.zeros((2, 2))), promotes=["*"])
        model.add_subsystem("comp", examp(), promotes=["*"])
        model.add_subsystem(
            "con",
            om.ExecComp("c = areas - 20.0", c=np.zeros((2, 2)), areas=np.zeros((2, 2))),
            promotes=["*"],
        )
        model.add_subsystem(
            "obj",
            om.ExecComp("o = areas[0, 0]", areas=np.zeros((2, 2))),
            promotes=["*"],
        )

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("widths", lower=-50.0, upper=50.0)
        model.add_objective("o")
        model.add_constraint("c", equals=0.0)

        prob.setup()

        failed = prob.run_driver()

        obj = prob["o"]
        assert_near_equal(obj, 20.0, 1e-6)

    def test_simple_array_comp2D_eq_con(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("widths", np.zeros((2, 2))), promotes=["*"])
        model.add_subsystem("comp", examp(), promotes=["*"])
        model.add_subsystem(
            "obj",
            om.ExecComp("o = areas[0, 0] + areas[1, 1]", areas=np.zeros((2, 2))),
            promotes=["*"],
        )

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("widths", lower=-50.0, upper=50.0)
        model.add_objective("o")
        model.add_constraint("areas", equals=np.array([24.0, 21.0, 3.5, 17.5]))

        prob.setup()

        failed = prob.run_driver()

        obj = prob["o"]
        assert_near_equal(obj, 41.5, 1e-6)

    def test_simple_array_comp2D_dbl_sided_con(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("widths", np.zeros((2, 2))), promotes=["*"])
        model.add_subsystem("comp", examp(), promotes=["*"])
        model.add_subsystem(
            "obj",
            om.ExecComp("o = areas[0, 0]", areas=np.zeros((2, 2))),
            promotes=["*"],
        )

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("widths", lower=-50.0, upper=50.0)
        model.add_objective("o")
        model.add_constraint(
            "areas",
            lower=np.array([24.0, 21.0, 3.5, 17.5]),
            upper=np.array([24.0, 21.0, 3.5, 17.5]),
        )

        prob.setup()

        failed = prob.run_driver()

        con = prob["areas"]
        assert_near_equal(con, np.array([[24.0, 21.0], [3.5, 17.5]]), 1e-6)

    def test_simple_array_comp2D_dbl_sided_con_array(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("widths", np.zeros((2, 2))), promotes=["*"])
        model.add_subsystem("comp", examp(), promotes=["*"])
        model.add_subsystem(
            "obj",
            om.ExecComp("o = areas[0, 0]", areas=np.zeros((2, 2))),
            promotes=["*"],
        )

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("widths", lower=-50.0, upper=50.0)
        model.add_objective("o")
        model.add_constraint("areas", lower=20.0, upper=20.0)

        prob.setup()

        failed = prob.run_driver()

        obj = prob["o"]
        assert_near_equal(obj, 20.0, 1e-6)

    def test_simple_array_comp2D_array_lo_hi(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("widths", np.zeros((2, 2))), promotes=["*"])
        model.add_subsystem("comp", examp(), promotes=["*"])
        model.add_subsystem(
            "con",
            om.ExecComp("c = areas - 20.0", c=np.zeros((2, 2)), areas=np.zeros((2, 2))),
            promotes=["*"],
        )
        model.add_subsystem(
            "obj",
            om.ExecComp("o = areas[0, 0]", areas=np.zeros((2, 2))),
            promotes=["*"],
        )

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("widths", lower=-50.0 * np.ones((2, 2)), upper=50.0 * np.ones((2, 2)))
        model.add_objective("o")
        model.add_constraint("c", equals=0.0)

        prob.setup()

        failed = prob.run_driver()

        obj = prob["o"]
        assert_near_equal(obj, 20.0, 1e-6)

    def test_simple_paraboloid_scaled_desvars_fwd(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0, ref=0.02)
        model.add_design_var("y", lower=-50.0, upper=50.0, ref=0.02)
        model.add_objective("f_xy")
        model.add_constraint("c", lower=10.0, upper=11.0)

        prob.setup(check=False, mode="fwd")

        failed = prob.run_driver()

        assert_near_equal(prob["x"] - prob["y"], 11.0, 1e-6)

    def test_simple_paraboloid_scaled_desvars_rev(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0, ref=0.02)
        model.add_design_var("y", lower=-50.0, upper=50.0, ref=0.02)
        model.add_objective("f_xy")
        model.add_constraint("c", lower=10.0, upper=11.0)

        prob.setup(check=False, mode="rev")

        failed = prob.run_driver()

        assert_near_equal(prob["x"] - prob["y"], 11.0, 1e-6)

    def test_simple_paraboloid_scaled_constraint_fwd(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-8

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", lower=10.0, upper=11.0, ref=10.0)

        prob.setup(check=False, mode="fwd")

        failed = prob.run_driver()

        assert_near_equal(prob["x"] - prob["y"], 11.0, 1e-6)

    def test_simple_paraboloid_scaled_objective_fwd(self):
        prob = om.Problem(reports=False)
        model = prob.model

        prob.set_solver_print(level=0)

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy", ref=10.0)
        model.add_constraint("c", lower=10.0, upper=11.0)

        prob.setup(check=False, mode="fwd")

        failed = prob.run_driver()

        assert_near_equal(prob["x"] - prob["y"], 11.0, 1e-6)

    def test_simple_paraboloid_scaled_objective_rev(self):
        prob = om.Problem(reports=False)
        model = prob.model

        prob.set_solver_print(level=0)

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy", ref=10.0)
        model.add_constraint("c", lower=10.0, upper=11.0)

        prob.setup(check=False, mode="rev")

        failed = prob.run_driver()

        assert_near_equal(prob["x"] - prob["y"], 11.0, 1e-6)

    def test_sellar_mdf(self):
        prob = om.Problem(reports=False)
        model = prob.model = SellarDerivativesGrouped()

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("z", lower=np.array([-10.0, 0.0]), upper=np.array([10.0, 10.0]))
        model.add_design_var("x", lower=0.0, upper=10.0)
        model.add_objective("obj")
        model.add_constraint("con1", upper=0.0)
        model.add_constraint("con2", upper=0.0)

        prob.setup(check=False, mode="rev")

        failed = prob.run_driver()

        assert_near_equal(prob["z"][0], 1.9776, 1e-3)
        assert_near_equal(prob["z"][1], 0.0, 1e-3)
        assert_near_equal(prob["x"], 0.0, 1e-3)

    def test_bug_in_eq_constraints(self):
        # We were getting extra constraints created because lower and upper are maxfloat instead of
        # None when unused.
        p = om.Problem(reports=False, model=SineFitter())
        p.driver = NLoptDriver()

        p.setup()
        p.run_driver()

        max_defect = np.max(np.abs(p["defect.defect"]))
        assert_near_equal(max_defect, 0.0, 1e-10)

    def test_reraise_exception_from_callbacks(self):
        class ReducedActuatorDisc(om.ExplicitComponent):
            def setup(self):
                # Inputs
                self.add_input("a", 0.5, desc="Induced Velocity Factor")
                self.add_input(
                    "Vu",
                    10.0,
                    units="m/s",
                    desc="Freestream air velocity, upstream of rotor",
                )

                # Outputs
                self.add_output(
                    "Vd",
                    0.0,
                    units="m/s",
                    desc="Slipstream air velocity, downstream of rotor",
                )

            def compute(self, inputs, outputs):
                a = inputs["a"]
                Vu = inputs["Vu"]

                outputs["Vd"] = Vu * (1 - 2 * a)

            def compute_partials(self, inputs, J):
                Vu = inputs["Vu"]

                J["Vd", "a"] = -2.0 * Vu

        prob = om.Problem(reports=False)
        indeps = prob.model.add_subsystem("indeps", om.IndepVarComp(), promotes=["*"])
        indeps.add_output("a", 0.5)
        indeps.add_output("Vu", 10.0, units="m/s")

        prob.model.add_subsystem("a_disk", ReducedActuatorDisc(), promotes_inputs=["a", "Vu"])

        # setup the optimization
        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"

        prob.model.add_design_var("a", lower=0.0, upper=1.0)
        # negative one so we maximize the objective
        prob.model.add_objective("a_disk.Vd", scaler=-1)

        prob.setup()

        with self.assertRaises(KeyError) as context:
            prob.run_driver()

        msg = 'Variable name pair ("Vd", "a") must first be declared.'
        self.assertTrue(msg in str(context.exception))

    def test_simple_paraboloid_upper_LN_COBYLA(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LN_COBYLA"
        prob.driver.options["tol"] = 1e-12

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", upper=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        assert_near_equal(prob["x"], 7.16667, 1e-6)
        assert_near_equal(prob["y"], -7.833334, 1e-6)

    def test_sellar_mdf_LN_COBYLA(self):
        prob = om.Problem(reports=False)
        model = prob.model = SellarDerivativesGrouped()

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LN_COBYLA"
        prob.driver.options["tol"] = 1e-12

        prob.set_solver_print(level=0)

        model.add_design_var("z", lower=np.array([-10.0, 0.0]), upper=np.array([10.0, 10.0]))
        model.add_design_var("x", lower=0.0, upper=10.0)
        model.add_objective("obj")
        model.add_constraint("con1", upper=0.0)
        model.add_constraint("con2", upper=0.0)

        prob.setup(check=False, mode="rev")

        failed = prob.run_driver()

        assert_near_equal(prob["z"][0], 1.9776, 1e-3)
        assert_near_equal(prob["z"][1], 0.0, 1e-3)
        assert_near_equal(prob["x"], 0.0, 1e-2)

    def test_simple_paraboloid_lower_linear(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = x - y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", lower=15.0, linear=True)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        assert_near_equal(prob["x"], 7.16667, 1e-6)
        assert_near_equal(prob["y"], -7.833334, 1e-6)

        self.assertEqual(prob.driver._obj_and_nlcons, ["f_xy"])

    def test_simple_paraboloid_equality_linear(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0, linear=True)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        assert_near_equal(prob["x"], 7.16667, 1e-6)
        assert_near_equal(prob["y"], -7.833334, 1e-6)

    def test_debug_print_option(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        prob.driver.options["debug_print"] = ["desvars", "ln_cons", "nl_cons", "objs"]

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", upper=-15.0)

        prob.setup()

        failed, output = run_driver(prob)

        output = output.split("\n")

        self.assertTrue(
            output.count("Design Vars") > 1,
            "Should be more than one design vars header printed",
        )
        self.assertTrue(
            output.count("Nonlinear constraints") > 1,
            "Should be more than one nonlinear constraint header printed",
        )
        self.assertTrue(
            output.count("Linear constraints") > 1,
            "Should be more than one linear constraint header printed",
        )
        self.assertTrue(
            output.count("Objectives") > 1,
            "Should be more than one objective header printed",
        )
        self.assertTrue(
            len([s for s in output if s.startswith("{'x")]) > 1,
            "Should be more than one p1.x printed",
        )
        self.assertTrue(
            len([s for s in output if "'y'" in s]) > 1,
            "Should be more than one p2.y printed",
        )
        self.assertTrue(
            len([s for s in output if s.startswith("{'c")]) > 1,
            "Should be more than one con.c printed",
        )
        self.assertTrue(
            len([s for s in output if s.startswith("{'f_xy")]) > 1,
            "Should be more than one comp.f_xy printed",
        )

    def test_sellar_mdf_linear_con_directsolver(self):
        # This test makes sure that we call solve_nonlinear first if we have any linear constraints
        # to cache.
        prob = om.Problem(reports=False)
        model = prob.model = SellarDerivatives()

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-3

        model.add_design_var("z", lower=np.array([-10.0, 0.0]), upper=np.array([10.0, 10.0]))
        model.add_design_var("x", lower=0.0, upper=10.0)
        model.add_objective("obj")
        model.add_constraint("con1", upper=0.0)
        model.add_constraint("con2", upper=0.0)
        model.add_constraint("x", upper=11.0, linear=True)

        prob.setup(check=False, mode="rev")
        prob.set_solver_print(level=0)

        failed = prob.run_driver()

        assert_near_equal(prob["z"][0], 1.9776, 1e-3)
        assert_near_equal(prob["z"][1], 0.0, 1e-3)
        assert_near_equal(prob["x"], 0.0, 4e-3)

        self.assertEqual(len(prob.driver._lincongrad_cache), 1)
        # Piggyback test: make sure we can run the driver again as a subdriver without a keyerror.
        prob.driver.run()
        self.assertEqual(len(prob.driver._lincongrad_cache), 1)

    def test_call_final_setup(self):
        # Make sure we call final setup if our model hasn't been setup.

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        with self.assertRaises(RuntimeError) as cm:
            totals = prob.check_totals(method="fd", out_stream=False)

        # expected_msg = "Problem: run_model must be called before total derivatives can be checked."
        expected_msg = "run_model must be called before total derivatives can be checked."

        # self.assertEqual(expected_msg, str(cm.exception))
        self.assertTrue(str(cm.exception).find(expected_msg) >= 0)

    def test_LN_COBYLA_linear_constraint(self):
        # Bug where NLoptDriver tried to compute and cache the constraint derivatives for the
        # lower and upper bounds of the desvars even though we were using a non-gradient optimizer.
        # This causd a KeyError.
        prob = om.Problem(reports=False)
        indeps = prob.model.add_subsystem("indeps", om.IndepVarComp())
        indeps.add_output("x", 3.0)
        indeps.add_output("y", -4.0)

        prob.model.add_subsystem("parab", Paraboloid())

        prob.model.add_subsystem("const", om.ExecComp("g = x + y"))

        prob.model.connect("indeps.x", ["parab.x", "const.x"])
        prob.model.connect("indeps.y", ["parab.y", "const.y"])

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LN_COBYLA"

        prob.model.add_constraint("const.g", lower=0, upper=10.0)
        prob.model.add_design_var("indeps.x", **{"ref0": 0, "ref": 2, "lower": -50, "upper": 50})
        prob.model.add_design_var("indeps.y", **{"ref0": 0, "ref": 2, "lower": -50, "upper": 50})
        prob.model.add_objective("parab.f_xy", scaler=4.0)
        prob.setup()
        prob.run_driver()

        # minimum value
        assert_near_equal(prob["parab.f_xy"], -27, 1e-6)

    def test_multiple_objectives_error(self):
        import openmdao.api as om
        from openmdao.test_suite.components.paraboloid import Paraboloid

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        self.assertFalse(prob.driver.supports["multiple_objectives"])
        prob.driver.options["debug_print"] = ["nl_cons", "objs"]

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_objective("c")  # Second objective
        prob.setup()

        with self.assertRaises(RuntimeError):
            prob.run_model()

        with self.assertRaises(RuntimeError):
            prob.run_driver()

    def test_GN_DIRECT(self):
        import openmdao.api as om

        np.random.seed(6)

        size = 2  # size of the design variable

        class Rastrigin(om.ExplicitComponent):
            def setup(self):
                self.add_input("x", 0.5 * np.ones(size))
                self.add_output("f", 0.5)

            def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
                x = inputs["x"]
                outputs["f"] = rastrigin(x)

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("indeps", om.IndepVarComp("x", np.ones(size)), promotes=["*"])
        model.add_subsystem("rastrigin", Rastrigin(), promotes=["*"])

        prob.driver = driver = NLoptDriver()
        driver.options["optimizer"] = "GN_DIRECT"
        driver.options["maxiter"] = 5000

        model.add_design_var("x", lower=-5.12 * np.ones(size), upper=5.12 * np.ones(size))
        model.add_objective("f")
        prob.setup()
        prob.run_driver()
        assert_near_equal(prob["x"], np.zeros(size), 1e-3)
        assert_near_equal(prob["f"], 0.0, 1e-5)

    def test_GN_DIRECT_L(self):
        import openmdao.api as om

        np.random.seed(6)

        size = 2  # size of the design variable

        class Rastrigin(om.ExplicitComponent):
            def setup(self):
                self.add_input("x", 0.5 * np.ones(size))
                self.add_output("f", 0.5)

            def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
                x = inputs["x"]
                outputs["f"] = rastrigin(x)

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("indeps", om.IndepVarComp("x", np.ones(size)), promotes=["*"])
        model.add_subsystem("rastrigin", Rastrigin(), promotes=["*"])

        prob.driver = driver = NLoptDriver()
        driver.options["optimizer"] = "GN_DIRECT_L"
        driver.options["maxiter"] = 500

        model.add_design_var("x", lower=-5.12 * np.ones(size), upper=5.12 * np.ones(size))
        model.add_objective("f")
        prob.setup()
        prob.run_driver()
        assert_near_equal(prob["x"], np.zeros(size), 1e-6)
        assert_near_equal(prob["f"], 0.0, 1e-6)

    def test_GN_DIRECT_L_NOSCAL(self):
        import openmdao.api as om

        np.random.seed(6)

        size = 2  # size of the design variable

        class Rastrigin(om.ExplicitComponent):
            def setup(self):
                self.add_input("x", 0.5 * np.ones(size))
                self.add_output("f", 0.5)

            def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
                x = inputs["x"]
                outputs["f"] = rastrigin(x)

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("indeps", om.IndepVarComp("x", np.ones(size)), promotes=["*"])
        model.add_subsystem("rastrigin", Rastrigin(), promotes=["*"])

        prob.driver = driver = NLoptDriver()
        driver.options["optimizer"] = "GN_DIRECT_L_NOSCAL"
        driver.options["maxiter"] = 500

        model.add_design_var("x", lower=-5.12 * np.ones(size), upper=5.12 * np.ones(size))
        model.add_objective("f")
        prob.setup()
        prob.run_driver()
        assert_near_equal(prob["x"], np.zeros(size), 1e-6)
        assert_near_equal(prob["f"], 0.0, 1e-6)

    def test_GN_ORIG_DIRECT(self):
        import openmdao.api as om

        np.random.seed(6)

        size = 2  # size of the design variable

        class Rastrigin(om.ExplicitComponent):
            def setup(self):
                self.add_input("x", 0.5 * np.ones(size))
                self.add_output("f", 0.5)

            def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
                x = inputs["x"]
                outputs["f"] = rastrigin(x)

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("indeps", om.IndepVarComp("x", np.ones(size)), promotes=["*"])
        model.add_subsystem("rastrigin", Rastrigin(), promotes=["*"])

        prob.driver = driver = NLoptDriver()
        driver.options["optimizer"] = "GN_ORIG_DIRECT"
        driver.options["maxiter"] = 1000

        model.add_design_var("x", lower=-5.12 * np.ones(size), upper=5.12 * np.ones(size))
        model.add_objective("f")
        prob.setup()
        prob.run_driver()
        assert_near_equal(prob["x"], np.zeros(size), 1e-5)
        assert_near_equal(prob["f"], 0.0, 1e-5)

    def test_GN_ORIG_DIRECT_L(self):
        import openmdao.api as om

        np.random.seed(6)

        size = 2  # size of the design variable

        class Rastrigin(om.ExplicitComponent):
            def setup(self):
                self.add_input("x", 0.5 * np.ones(size))
                self.add_output("f", 0.5)

            def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
                x = inputs["x"]
                outputs["f"] = rastrigin(x)

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("indeps", om.IndepVarComp("x", np.ones(size)), promotes=["*"])
        model.add_subsystem("rastrigin", Rastrigin(), promotes=["*"])

        prob.driver = driver = NLoptDriver()
        driver.options["optimizer"] = "GN_ORIG_DIRECT_L"
        driver.options["maxiter"] = 500

        model.add_design_var("x", lower=-5.12 * np.ones(size), upper=5.12 * np.ones(size))
        model.add_objective("f")
        prob.setup()
        prob.run_driver()
        assert_near_equal(prob["x"], np.zeros(size), 1e-6)
        assert_near_equal(prob["f"], 0.0, 1e-6)

    def test_simple_paraboloid_upper_LD_MMA(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_MMA"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", upper=-15.0)

        prob.setup()

        failed = prob.run_driver()

        assert_near_equal(prob["x"], 7.1666666, 1e-6)
        assert_near_equal(prob["y"], -7.83333333, 1e-6)

    def test_simple_paraboloid_upper_LD_CCSAQ(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_CCSAQ"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", upper=-15.0)

        prob.setup()

        failed = prob.run_driver()

        assert_near_equal(prob["x"], 7.1666666, 1e-6)
        assert_near_equal(prob["y"], -7.83333333, 1e-6)

    def test_simple_paraboloid_upper_GN_ORIG_DIRECT(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.2), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "GN_ORIG_DIRECT"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 500

        model.add_design_var("x", lower=5.0, upper=10.0)
        model.add_design_var("y", lower=-10.0, upper=-5.0)
        model.add_objective("f_xy")
        model.add_constraint("c", upper=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Just get pretty close to the optimum
        assert_near_equal(prob["x"], 7.1666666, 1e-2)
        assert_near_equal(prob["y"], -7.83333333, 1e-2)

    def test_simple_paraboloid_upper_GN_ORIG_DIRECT_L(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.2), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "GN_ORIG_DIRECT_L"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 500

        model.add_design_var("x", lower=5.0, upper=10.0)
        model.add_design_var("y", lower=-10.0, upper=-5.0)
        model.add_objective("f_xy")
        model.add_constraint("c", upper=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Just get pretty close to the optimum
        assert_near_equal(prob["x"], 7.1666666, 1e-2)
        assert_near_equal(prob["y"], -7.83333333, 1e-2)

    def test_simple_paraboloid_equality_COBYLA(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LN_COBYLA"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        assert_near_equal(prob["x"], 7.16667, 1e-4)
        assert_near_equal(prob["y"], -7.833334, 1e-4)

    def test_simple_paraboloid_equality_ISRES(self):
        prob = om.Problem(reports=False)
        model = prob.model

        # Start very close to the correct answer with tight bounds to
        # reduce test runtime
        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.6), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "GN_ISRES"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 10000

        model.add_design_var("x", lower=7.0, upper=7.5)
        model.add_design_var("y", lower=-8.0, upper=-7.5)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        # Loose tolerance
        assert_near_equal(prob["x"], 7.16667, 0.05)
        assert_near_equal(prob["y"], -7.833334, 0.05)

    def test_simple_paraboloid_equality_failure_MMA(self):
        prob = om.Problem(reports=False)
        model = prob.model

        # Start very close to the correct answer with tight bounds to
        # reduce test runtime
        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.6), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_MMA"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 5000

        model.add_design_var("x", lower=7.0, upper=7.5)
        model.add_design_var("y", lower=-8.0, upper=-7.5)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        with self.assertRaises(NotImplementedError) as raises_msg:
            failed = prob.run_driver()

        exception = raises_msg.exception

        msg = "The selected optimizer, LD_MMA, does not support equality constraints."

        self.assertIn(msg, exception.args[0])

    def test_simple_paraboloid_equality_failure_LD_CCSAQ(self):
        prob = om.Problem(reports=False)
        model = prob.model

        # Start very close to the correct answer with tight bounds to
        # reduce test runtime
        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.6), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_CCSAQ"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 5000

        model.add_design_var("x", lower=7.0, upper=7.5)
        model.add_design_var("y", lower=-8.0, upper=-7.5)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        with self.assertRaises(NotImplementedError) as raises_msg:
            failed = prob.run_driver()

        exception = raises_msg.exception

        msg = "The selected optimizer, LD_CCSAQ, does not support equality constraints."

        self.assertIn(msg, exception.args[0])

    def test_simple_paraboloid_equality_failure_GN_ORIG_DIRECT(self):
        prob = om.Problem(reports=False)
        model = prob.model

        # Start very close to the correct answer with tight bounds to
        # reduce test runtime
        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.6), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "GN_ORIG_DIRECT"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 5000

        model.add_design_var("x", lower=7.0, upper=7.5)
        model.add_design_var("y", lower=-8.0, upper=-7.5)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        with self.assertRaises(NotImplementedError) as raises_msg:
            failed = prob.run_driver()

        exception = raises_msg.exception

        msg = "The selected optimizer, GN_ORIG_DIRECT, does not support equality constraints."

        self.assertIn(msg, exception.args[0])

    def test_simple_paraboloid_equality_failure_GN_ORIG_DIRECT_L(self):
        prob = om.Problem(reports=False)
        model = prob.model

        # Start very close to the correct answer with tight bounds to
        # reduce test runtime
        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.6), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "GN_ORIG_DIRECT_L"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 5000

        model.add_design_var("x", lower=7.0, upper=7.5)
        model.add_design_var("y", lower=-8.0, upper=-7.5)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        with self.assertRaises(NotImplementedError) as raises_msg:
            failed = prob.run_driver()

        exception = raises_msg.exception

        msg = "The selected optimizer, GN_ORIG_DIRECT_L, does not support equality constraints."

        self.assertIn(msg, exception.args[0])

    def test_simple_paraboloid_equality_failure_GN_AGS(self):
        prob = om.Problem(reports=False)
        model = prob.model

        # Start very close to the correct answer with tight bounds to
        # reduce test runtime
        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.6), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "GN_AGS"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 5000

        model.add_design_var("x", lower=7.0, upper=7.5)
        model.add_design_var("y", lower=-8.0, upper=-7.5)
        model.add_objective("f_xy")
        model.add_constraint("c", equals=-15.0)

        prob.setup()

        with self.assertRaises(NotImplementedError) as raises_msg:
            failed = prob.run_driver()

        exception = raises_msg.exception

        msg = "The selected optimizer, GN_AGS, does not support equality constraints."

        self.assertIn(msg, exception.args[0])

    def test_maxtime(self):
        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver(optimizer="LD_SLSQP", tol=1e-9, maxtime=0.0001)

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        failed = prob.run_driver()

        # It shouldn't have time to move from the initial point
        assert_near_equal(prob["x"], 50.0, 1e-6)
        assert_near_equal(prob["y"], 50.0, 1e-6)

    def test_simple_paraboloid_inequalities(self):
        # This test checks that you can set a constraint with the same value
        # for both the lower and upper bounds to effectively create an equality
        # constraint even if the optimization method doesn't allow equality constraints

        prob = om.Problem(reports=False)
        model = prob.model

        # Start very close to the correct answer with tight bounds to
        # reduce test runtime
        model.add_subsystem("p1", om.IndepVarComp("x", 7.5), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", -7.6), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])
        model.add_subsystem("con", om.ExecComp("c = - x + y"), promotes=["*"])

        prob.set_solver_print(level=0)

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "GN_ORIG_DIRECT"
        prob.driver.options["tol"] = 1e-6

        prob.driver.options["maxiter"] = 5000

        model.add_design_var("x", lower=7.0, upper=7.5)
        model.add_design_var("y", lower=-8.0, upper=-7.5)
        model.add_objective("f_xy")
        model.add_constraint("c", lower=-15.0, upper=-15.0)

        prob.setup()

        failed = prob.run_driver()

        # Minimum should be at (7.166667, -7.833334)
        # Loose tolerance
        assert_near_equal(prob["x"], 7.16667, 1e-2)
        assert_near_equal(prob["y"], -7.833334, 1e-2)


@unittest.skipIf(nlopt is None, "only run if NLopt is installed.")
class TestNLoptDriverFeatures(unittest.TestCase):
    def test_feature_basic(self):
        import openmdao.api as om
        from openmdao.test_suite.components.paraboloid import Paraboloid

        from wisdem.optimization_drivers.nlopt_driver import NLoptDriver

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.driver = NLoptDriver()
        prob.driver.options["optimizer"] = "LD_SLSQP"
        prob.driver.options["tol"] = 1e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        prob.run_driver()

        assert_near_equal(prob["x"], 6.66666667, 1e-6)
        assert_near_equal(prob["y"], -7.3333333, 1e-6)

    def test_feature_optimizer(self):
        import openmdao.api as om
        from openmdao.test_suite.components.paraboloid import Paraboloid

        from wisdem.optimization_drivers.nlopt_driver import NLoptDriver

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.driver = NLoptDriver(optimizer="LD_SLSQP")

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        prob.run_driver()

        assert_near_equal(prob["x"], 6.66666667, 1e-6)
        assert_near_equal(prob["y"], -7.3333333, 1e-6)

    def test_feature_maxiter(self):
        import openmdao.api as om
        from openmdao.test_suite.components.paraboloid import Paraboloid

        from wisdem.optimization_drivers.nlopt_driver import NLoptDriver

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.driver = NLoptDriver()
        prob.driver.options["maxiter"] = 20

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        prob.run_driver()

        assert_near_equal(prob["x"], 6.66666667, 1e-6)
        assert_near_equal(prob["y"], -7.3333333, 1e-6)

    def test_feature_tol(self):
        import openmdao.api as om
        from openmdao.test_suite.components.paraboloid import Paraboloid

        from wisdem.optimization_drivers.nlopt_driver import NLoptDriver

        prob = om.Problem(reports=False)
        model = prob.model

        model.add_subsystem("p1", om.IndepVarComp("x", 50.0), promotes=["*"])
        model.add_subsystem("p2", om.IndepVarComp("y", 50.0), promotes=["*"])
        model.add_subsystem("comp", Paraboloid(), promotes=["*"])

        prob.driver = NLoptDriver()
        prob.driver.options["tol"] = 1.0e-9

        model.add_design_var("x", lower=-50.0, upper=50.0)
        model.add_design_var("y", lower=-50.0, upper=50.0)
        model.add_objective("f_xy")

        prob.setup()

        prob.run_driver()

        assert_near_equal(prob["x"], 6.66666667, 1e-6)
        assert_near_equal(prob["y"], -7.3333333, 1e-6)


@unittest.skipUnless(nlopt is None, "only run if NLopt is NOT installed.")
class TestNotInstalled(unittest.TestCase):
    def test_nlopt_not_installed(self):
        # the import should not fail
        from wisdem.optimization_drivers.nlopt_driver import NLoptDriver

        # but we get a RuntimeError if we try to instantiate
        with self.assertRaises(RuntimeError) as ctx:
            NLoptDriver()

        self.assertEqual(str(ctx.exception), "NLoptDriver is not available, NLopt is not installed.")


if __name__ == "__main__":
    unittest.main()



import os
import copy
import numpy as np
import jsonschema as json
import ruamel.yaml as ry
from functools import reduce
import operator


fschema_geom = os.path.join(os.path.dirname(os.path.realpath(__file__)), "geometry_schema.yaml")
fschema_model = os.path.join(os.path.dirname(os.path.realpath(__file__)), "modeling_schema.yaml")
fschema_opt = os.path.join(os.path.dirname(os.path.realpath(__file__)), "analysis_schema.yaml")


def load_yaml(fname_input : str) -> dict:
    """
    Reads and parses a YAML file in a safe mode using the ruamel.yaml library.

    Args:
        fname_input (str): Path to the YAML file to be loaded.

    Returns:
        dict: Parsed YAML content as a dictionary.
    """
    reader = ry.YAML(typ="safe", pure=True)
    with open(fname_input, "r", encoding="utf-8") as f:
        input_yaml = reader.load(f)
    return input_yaml

def write_yaml(instance : dict, foutput : str) -> None:
    """
    Writes a dictionary to a YAML file using the ruamel.yaml library.

    Args:
        instance (dict): Dictionary to be written to the YAML file.
        foutput (str): Path to the output YAML file.

    Returns:
        None
    """
    instance = remove_numpy(instance)

    # Write yaml with updated values
    yaml = ry.YAML()
    yaml.default_flow_style = None
    yaml.width = float("inf")
    yaml.indent(mapping=4, sequence=6, offset=3)
    yaml.allow_unicode = False
    with open(foutput, "w", encoding="utf-8") as f:
        yaml.dump(instance, f)


# ---------------------
# This is for when the defaults are in another file
def nested_get(indict, keylist):
    rv = indict
    for k in keylist:
        rv = rv[k]
    return rv


def nested_set(indict, keylist, val):
    rv = indict
    for i, k in enumerate(keylist):
        rv = rv[k] if i != len(keylist) - 1 else val


def integrate_defaults(instance : dict, defaults : dict, yaml_schema : dict) -> dict:
    """
    Integrates default values from a dictionary into another dictionary.

    Args:
        instance (dict): Dictionary to be updated with default values.
        defaults (dict): Dictionary containing default values.
        yaml_schema (dict): Dictionary containing the schema of the YAML file.

    Returns:
        dict: Updated dictionary with default values integrated.
    """
    # Prep iterative validator
    # json.validate(self.wt_init, yaml_schema)
    validator = json.Draft7Validator(yaml_schema)
    errors = validator.iter_errors(instance)

    # Loop over errors
    for e in errors:
        # If the error is due to a missing required value, try to set it to the default
        if e.validator == "required":
            for k in e.validator_value:
                if k not in e.instance.keys():
                    mypath = e.absolute_path.copy()
                    mypath.append(k)
                    v = nested_get(defaults, mypath)
                    if isinstance(v, dict) or isinstance(v, list) or v in ["name", "material"]:
                        # Too complicated to just copy over default, so give it back to the user
                        raise (e)
                    print("WARNING: Missing value,", list(mypath), ", so setting to:", v)
                    nested_set(instance, mypath, v)
        raise (e)
    return instance


def simple_types(indict : dict) -> dict:
    """
    Recursively converts numpy array elements within a nested dictionary to lists and ensures
    all values are simple types (float, int, dict, bool, str).

    Args:
        indict (dict): The dictionary to process.

    Returns:
        dict: The processed dictionary with numpy arrays converted to lists and unsupported types to empty strings.
    """
    def convert(value):
        if isinstance(value, np.ndarray):
            return value.tolist()
        elif isinstance(value, dict):
            return simple_types(value)
        elif isinstance(value, (list, tuple)):
            return [convert(item) for item in value]
        elif isinstance(value, (float, int, bool, str)):
            return value
        else:
            return ""

    rv = {k: convert(v) for k, v in indict.items()}
    return rv


# ---------------------
# See: https://python-jsonschema.readthedocs.io/en/stable/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance
def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])

        for error in validate_properties(validator, properties, instance, schema):
            yield error

    return json.validators.extend(validator_class, {"properties": set_defaults})


DefaultValidatingDraft7Validator = extend_with_default(json.Draft7Validator)

def _validate(finput, fschema, defaults=True):
    """
    Validates a dictionary against a schema and returns the validated dictionary.

    Args:
        finput (dict or str): Dictionary or path to the YAML file to be validated.
        fschema (dict or str): Dictionary or path to the schema file to validate against.
        defaults (bool): Flag to indicate if default values should be integrated.

    Returns:
        dict: Validated dictionary.
    """
    schema_dict = fschema if isinstance(fschema, dict) else load_yaml(fschema)
    input_dict = finput if isinstance(finput, dict) else load_yaml(finput)
    validator = DefaultValidatingDraft7Validator if defaults else json.Draft7Validator
    validator(schema_dict).validate(input_dict)
    return input_dict


# ---------------------
def load_geometry_yaml(finput):
    return _validate(finput, fschema_geom)


def load_modeling_yaml(finput):
    return _validate(finput, fschema_model)


def load_analysis_yaml(finput):
    return _validate(finput, fschema_opt)


def write_geometry_yaml(instance : dict, foutput : str) -> None:
    _validate(instance, fschema_geom, defaults=False)
    # Ensure the file output has a .yaml suffix
    if not foutput.endswith(".yaml"):
        foutput += ".yaml"
    write_yaml(instance, foutput)


def write_modeling_yaml(instance : dict, foutput : str) -> None:
    _validate(instance, fschema_model, defaults=False)

    # Ensure the output filename does not end with .yaml or .yml
    if foutput.endswith(".yaml"):
        foutput = foutput[:-5]
    elif foutput.endswith(".yml"):
        foutput = foutput[:-4]
    sfx_str = "-modeling.yaml"

    instance2 = simple_types(instance)
    write_yaml(instance2, foutput + sfx_str)


def write_analysis_yaml(instance : dict, foutput : str) -> None:
    _validate(instance, fschema_opt, defaults=False)

    # Ensure the output filename does not end with .yaml or .yml
    if foutput.endswith(".yaml"):
        foutput = foutput[:-5]
    elif foutput.endswith(".yml"):
        foutput = foutput[:-4]

    sfx_str = "-analysis.yaml"
    write_yaml(instance, foutput + sfx_str)

def remove_numpy(fst_vt : dict) -> dict:
    """
    Recursively converts numpy array elements within a nested dictionary to lists and ensures
    all values are simple types (float, int, dict, bool, str) for writing to a YAML file.

    Args:
        fst_vt (dict): The dictionary to process.

    Returns:
        dict: The processed dictionary with numpy arrays converted to lists and unsupported types to simple types.
    """

    def get_dict(vartree, branch):
        return reduce(operator.getitem, branch, vartree)

    # Define conversion dictionary for numpy types
    conversions = {
        np.int_: int,
        np.intc: int,
        np.intp: int,
        np.int8: int,
        np.int16: int,
        np.int32: int,
        np.int64: int,
        np.uint8: int,
        np.uint16: int,
        np.uint32: int,
        np.uint64: int,
        np.single: float,
        np.double: float,
        np.longdouble: float,
        np.csingle: float,
        np.cdouble: float,
        np.float_: float,
        np.float16: float,
        np.float32: float,
        np.float64: float,
        np.complex64: float,
        np.complex128: float,
        np.bool_: bool,
        np.ndarray: lambda x: x.tolist(),
    }

    def loop_dict(vartree, branch):
        if not isinstance(vartree, dict):
            return fst_vt
        for var in vartree.keys():
            branch_i = copy.copy(branch)
            branch_i.append(var)
            if isinstance(vartree[var], dict):
                loop_dict(vartree[var], branch_i)
            else:
                current_value = get_dict(fst_vt, branch_i[:-1])[branch_i[-1]]
                data_type = type(current_value)
                if data_type in conversions:
                    get_dict(fst_vt, branch_i[:-1])[branch_i[-1]] = conversions[data_type](current_value)
                elif isinstance(current_value, (list, tuple)):
                    for i, item in enumerate(current_value):
                        current_value[i] = remove_numpy(item)

    # set fast variables to update values
    loop_dict(fst_vt, [])
    return fst_vt


if __name__ == "__main__":
    yaml_schema = load_yaml(fschema_opt)
    myobj = load_yaml("sample_analysis.yaml")
    DefaultValidatingDraft7Validator(yaml_schema).validate(myobj)
    # validator.validate( myobj )
    print([k for k in myobj.keys()])
    print(myobj["general"])

    obj = {}
    schema = {"properties": {"foo": {"default": "bar"}}}
    # Note jsonschem.validate(obj, schema, cls=DefaultValidatingDraft7Validator)
    # will not work because the metaschema contains `default` directives.
    DefaultValidatingDraft7Validator(schema).validate(obj)
    print(obj)



"""`MonopileInstallation` class and related processes."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import numpy as np
import simpy
from marmot import process

from wisdem.orbit.core import SubstructureDelivery
from wisdem.orbit.core.logic import (
    prep_for_site_operations,
    shuttle_items_to_queue_wait,
    get_list_of_items_from_port_wait,
)
from wisdem.orbit.phases.install import InstallPhase
from wisdem.orbit.core.exceptions import ItemNotFound

from .common import Monopile, TransitionPiece, upend_monopile, install_monopile, install_transition_piece


class MonopileInstallation(InstallPhase):
    """
    Standard monopile installation module using a Wind Turbine Installation
    Vessel (WTIV). If input `feeder` and `num_feeders` are not supplied, the
    WTIV will perform all transport and installation tasks. If the above inputs
    are defined, feeder barges will transport monopile components from port to
    site.
    """

    phase = "Monopile Installation"
    capex_category = "Substructure"

    #:
    expected_config = {
        "wtiv": "dict | str",
        "feeder": "dict | str (optional)",
        "num_feeders": "int (optional)",
        "site": {"depth": "m", "distance": "km"},
        "plant": {"num_turbines": "int"},
        "turbine": {"hub_height": "m"},
        "port": {
            "num_cranes": "int (optional, default: 1)",
            "monthly_rate": "USD/mo (optional)",
            "name": "str (optional)",
        },
        "monopile": {
            "length": "m",
            "diameter": "m",
            "deck_space": "m2",
            "mass": "t",
            "unit_cost": "USD",
        },
        "transition_piece": {
            "deck_space": "m2",
            "mass": "t",
            "unit_cost": "USD",
        },
        "monopile_supply_chain": {
            "enabled": "(optional, default: False)",
            "substructure_delivery_time": "h (optional, default: 168)",
            "num_substructures_delivered": "int (optional: default: 1)",
            "substructure_storage": "int (optional, default: inf)",
        },
    }

    def __init__(self, config, weather=None, **kwargs):
        """
        Creates an instance of MonopileInstallation.

        Parameters
        ----------
        config : dict
            Simulation specific configuration.
        weather : pd.DataFrame (optional)
            Weather profile at site.
            Expects columns 'waveheight' and 'windspeed'.
        """

        super().__init__(weather, **kwargs)

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.initialize_port()
        self.initialize_substructure_delivery()
        self.initialize_wtiv()
        self.setup_simulation(**kwargs)

    @property
    def system_capex(self):
        """Returns procurement cost of the substructures."""

        return (self.config["monopile"]["unit_cost"] + self.config["transition_piece"]["unit_cost"]) * self.config[
            "plant"
        ]["num_turbines"]

    def initialize_substructure_delivery(self):
        """ """

        monopile = Monopile(**self.config["monopile"])
        tp = TransitionPiece(**self.config["transition_piece"])
        self.set_mass = monopile.mass + tp.mass
        self.set_deck_space = monopile.deck_space + tp.deck_space

        self.num_monopiles = self.config["plant"]["num_turbines"]
        self.supply_chain = self.config.get("monopile_supply_chain", {})

        if self.supply_chain.get("enabled", False):
            delivery_time = self.supply_chain.get("substructure_delivery_time", 168)
            # storage = self.supply_chain.get("substructure_storage", "inf")
            supply_chain = SubstructureDelivery(
                "Monopile",
                self.num_monopiles,
                delivery_time,
                self.port,
                [monopile, tp],
                num_parallel=self.supply_chain.get("num_substructures_delivered", 1),
            )

            self.env.register(supply_chain)
            supply_chain.start()

        else:
            for _ in range(self.num_monopiles):
                self.port.put(monopile)
                self.port.put(tp)

    def setup_simulation(self, **kwargs):
        """
        Sets up simulation infrastructure, routing to specific methods dependent
        on number of feeders.
        """

        if self.config.get("num_feeders", None):
            self.initialize_feeders()
            self.initialize_queue()
            self.setup_simulation_with_feeders(**kwargs)

        else:
            self.feeders = None
            self.setup_simulation_without_feeders(**kwargs)

    def setup_simulation_without_feeders(self, **kwargs):
        """
        Sets up infrastructure for turbine installation without feeder barges.
        """

        site_distance = self.config["site"]["distance"]
        site_depth = self.config["site"]["depth"]
        hub_height = self.config["turbine"]["hub_height"]

        self.sets_per_trip = int(
            min(
                np.floor(self.wtiv.storage.max_cargo_mass / self.set_mass),
                np.floor(self.wtiv.storage.max_deck_space / self.set_deck_space),
            )
        )

        solo_install_monopiles(
            self.wtiv,
            port=self.port,
            distance=site_distance,
            monopiles=self.num_monopiles,
            site_depth=site_depth,
            hub_height=hub_height,
            per_trip=self.sets_per_trip,
            **kwargs,
        )

    def setup_simulation_with_feeders(self, **kwargs):
        """
        Sets up infrastructure for turbine installation using feeder barges.
        """

        site_distance = self.config["site"]["distance"]
        site_depth = self.config["site"]["depth"]
        component_list = ["Monopile", "TransitionPiece"]

        self.sets_per_trip = int(
            min(
                np.floor(self.feeders[0].storage.max_cargo_mass / self.set_mass),
                np.floor(self.feeders[0].storage.max_deck_space / self.set_deck_space),
            )
        )

        install_monopiles_from_queue(
            self.wtiv,
            queue=self.active_feeder,
            monopiles=self.num_monopiles,
            distance=site_distance,
            site_depth=site_depth,
            **kwargs,
        )

        assignments = [
            self.num_monopiles // len(self.feeders) + (1 if x < self.num_monopiles % len(self.feeders) else 0)
            for x in range(len(self.feeders))
        ]

        for assigned, feeder in zip(assignments, self.feeders):
            shuttle_items_to_queue_wait(
                feeder,
                port=self.port,
                queue=self.active_feeder,
                distance=site_distance,
                items=component_list,
                per_trip=self.sets_per_trip,
                assigned=assigned,
                **kwargs,
            )

    def initialize_wtiv(self):
        """
        Initializes the WTIV simulation object and the onboard vessel storage.
        """

        wtiv_specs = self.config.get("wtiv", None)
        name = wtiv_specs.get("name", "WTIV")

        wtiv = self.initialize_vessel(name, wtiv_specs)
        self.env.register(wtiv)

        wtiv.initialize()
        wtiv.at_port = True
        wtiv.at_site = False
        self.wtiv = wtiv

    def initialize_feeders(self):
        """
        Initializes feeder barge objects.
        """

        number = self.config.get("num_feeders", None)
        feeder_specs = self.config.get("feeder", None)

        self.feeders = []
        for n in range(number):
            name = "Feeder {}".format(n)

            feeder = self.initialize_vessel(name, feeder_specs)
            self.env.register(feeder)

            feeder.initialize()
            feeder.at_port = True
            feeder.at_site = False
            self.feeders.append(feeder)

    def initialize_queue(self):
        """
        Initializes the queue, modeled as a ``SimPy.Resource`` that feeders
        join at site. This limits the simulation to one active feeder at a time.
        """

        self.active_feeder = simpy.Resource(self.env, capacity=1)
        self.active_feeder.vessel = None
        self.active_feeder.activate = self.env.event()

    @property
    def detailed_output(self):
        """Returns detailed outputs of the monopile installation."""

        if self.feeders:
            transport_vessels = [*self.feeders]

        else:
            transport_vessels = [self.wtiv]

        outputs = {
            self.phase: {
                **self.agent_efficiencies,
                **self.get_max_cargo_mass_utilzations(transport_vessels),
                **self.get_max_deck_space_utilzations(transport_vessels),
            }
        }

        return outputs


@process
def solo_install_monopiles(vessel, port, distance, monopiles, **kwargs):
    """
    Logic that a Wind Turbine Installation Vessel (WTIV) uses during a single
    monopile installation process.

    Parameters
    ----------
    vessel : vessels.Vessel
        Vessel object that represents the WTIV.
    port : Port
    distance : int | float
        Distance between port and site (km).
    monopiles : int
        Total monopiles to install.
    """

    component_list = ["Monopile", "TransitionPiece"]

    n = 0
    while n < monopiles:
        if vessel.at_port:
            try:
                # Get substructure + transition piece from port
                yield get_list_of_items_from_port_wait(vessel, port, component_list, **kwargs)

            except ItemNotFound:
                # If no items are at port and vessel.storage.items is empty,
                # the job is done
                if not vessel.storage.items:
                    vessel.submit_debug_log(message="Item not found. Shutting down.")
                    break

            # Transit to site
            vessel.update_trip_data()
            vessel.at_port = False
            yield vessel.transit(distance)
            vessel.at_site = True

        if vessel.at_site:
            if vessel.storage.items:
                # Prep for monopile install
                yield prep_for_site_operations(vessel, survey_required=True, **kwargs)

                # Get monopile from internal storage
                monopile = yield vessel.get_item_from_storage("Monopile", **kwargs)

                yield upend_monopile(vessel, monopile.length, **kwargs)
                yield install_monopile(vessel, monopile, **kwargs)

                # Get transition piece from internal storage
                tp = yield vessel.get_item_from_storage("TransitionPiece", **kwargs)

                yield install_transition_piece(vessel, tp, **kwargs)
                vessel.submit_debug_log(progress="Substructure")
                n += 1

            else:
                # Transit to port
                vessel.at_site = False
                yield vessel.transit(distance)
                vessel.at_port = True

    vessel.submit_debug_log(message="Monopile installation complete!")


@process
def install_monopiles_from_queue(wtiv, queue, monopiles, distance, **kwargs):
    """
    Logic that a Wind Turbine Installation Vessel (WTIV) uses to install
    monopiles and transition pieces from queue of feeder barges.

    Parameters
    ----------
    env : simulation.Environment
        SimPy environment that the simulation runs in.
    wtiv : vessels.Vessel
        Vessel object that represents the WTIV.
    queue : simpy.Resource
        Queue object to interact with active feeder barge.
    number : int
        Total monopiles to install.
    distance : int | float
        Distance from site to port (km).
    """

    n = 0
    while n < monopiles:
        if wtiv.at_port:
            # Transit to site
            wtiv.at_port = False
            yield wtiv.transit(distance)
            wtiv.at_site = True

        if wtiv.at_site:
            if queue.vessel:
                # Prep for monopile install
                yield prep_for_site_operations(wtiv, survey_required=True, **kwargs)

                # Get monopile
                monopile = yield wtiv.get_item_from_storage("Monopile", vessel=queue.vessel, **kwargs)

                yield upend_monopile(wtiv, monopile.length, **kwargs)
                yield install_monopile(wtiv, monopile, **kwargs)

                # Get transition piece from active feeder
                tp = yield wtiv.get_item_from_storage(
                    "TransitionPiece",
                    vessel=queue.vessel,
                    release=True,
                    **kwargs,
                )

                # Install transition piece
                yield install_transition_piece(wtiv, tp, **kwargs)
                wtiv.submit_debug_log(progress="Substructure")
                n += 1

            else:
                start = wtiv.env.now
                yield queue.activate
                delay_time = wtiv.env.now - start
                wtiv.submit_action_log("Delay", delay_time, location="Site")

    # Transit to port
    wtiv.at_site = False
    yield wtiv.transit(distance)
    wtiv.at_port = True

    wtiv.submit_debug_log(message="Monopile installation complete!")



"""Tests suite for ORBIT."""

__author__ = ["Jake Nunemaker", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"



import traceback
import math

import pandas as pd
import numpy as np
from scipy.optimize import root_scalar

from wisdem.landbosse.model.WeatherDelay import WeatherDelay as WD
from wisdem.landbosse.model.CostModule import CostModule


class FoundationCost(CostModule):
    """
    **FoundationCost.py**

    - Created by Annika Eberle and Owen Roberts on Apr. 3, 2018

    - Refactored by Parangat Bhaskar and Alicia Key on June 3, 2019

    \nCalculates the costs of constructing foundations for land-based wind projects *(items in brackets are not yet implemented)*:

    * Get number of turbines
    * Get duration of construction
    * Get daily hours of operation*  # todo: add to process diagram
    * Get season of construction*  # todo: add to process diagram
    * [Get region]
    * Get rotor diameter
    * Get hub height
    * Get turbine rating
    * Get buoyant foundation design flag
    * [Get seismic zone]
    * Get tower technology type
    * Get hourly weather data
    * [Get specific seasonal delays]
    * [Get long-term, site-specific climate data]
    * Get price data
    * Get labor rates
    * Get material prices for steel and concrete
    * [Use region to determine weather data]


    \n\nGiven below is the set of calculations carried out in this module:

    * Calculate the foundation loads using the rotor diameter, hub height, and turbine rating

    * Determine the foundation size based on the foundation loads, buoyant foundation design flag, and type of tower technology

    * Estimate the amount of material needed for foundation construction based on foundation size and number of turbines

    * Estimate the amount of time required to construct foundation based on foundation size, hours of operation, duration of construction, and number of turbines

    * Estimate the additional amount of time for weather delays (currently only assessing wind delays) based on hourly weather data, construction time, hours of operation, and season of construction

    * Estimate the amount of labor required for foundation construction based on foundation size, construction time, and weather delay
        * Calculate number of workers by crew type
        * Calculate man hours by crew type

    * Estimate the amount of equipment needed for foundation construction based on foundation size, construction time, and weather delay
        * Calculate number of equipment by equip type
        * Calculate equipment hours by equip type

    - Calculate the total foundation cost based on amount of equipment, amount of labor, amount of material, and price data.


    **Keys in the input dictionary are the following:**

    depth
        (int) depth of foundation [in m]


    component_data
        (pd.DataFrame) data frame with wind turbine component data

    def __init__(self, input_dict, output_dict, project_name):
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name


    num_turbines
        (int) total number of turbines in wind farm

    duration_construction
        (int) estimated construction time in months

    num_delays
        (int) Number of delay events

    avg_hours_per_delay
        (float) Average hours per delay event

    std_dev_hours_per_delay
        (float) Standard deviation from average hours per delay event

    delay_speed_m_per_s
        (float) wind speed above which weather delays kick in

    start_delay_hours
        (int)

    mission_time_hours
        (int)

    gust_wind_speed_m_per_s
        (float)

    wind_height_of_interest_m
        (int)

    wind_shear_exponent
        (float)

    season_construct
        list of seasons (like ['spring', 'summer']) for the construction.

    time_construct
        list of time windows for constructions. Use ['normal'] for a
        0800 to 1800 schedule 10 hour schedule. Use ['long'] for an
        overnight 1800 to 2359, 0000 to 0759 overnight schedule. Use
        ['normal', 'long'] for a 24-hour schedule.

    operational_hrs_per_day
        (float)


    material_price
        (pd.DataFrame) dataframe containing foundation cost related material prices

    rsmeans
        (pd.DataFrame) TODO: Formal definition for rsmeans?


    **Keys in the output dictionary are the following:**

    F_dead_kN_per_turbine
        (float) foundation dead load [in kN]

    F_horiz_kN_per_turbine
        (float) total lateral load [kN]

    M_tot_kN_m_per_turbine
        (float) Moment [kN.m]

    Radius_o_m
        (float) foundation radius based on overturning moment [in m]

    Radius_g_m
        (float) foundation radius based on gapping [in m]

    Radius_b_m
        (float) foundation radius based on bearing pressure [in m]

    Radius_m
        (float) largest foundation radius based on all three foundation design criteria: moment, gapping, bearing [in m]

    foundation_volume_concrete_m3_per_turbine
        (float) volume of a round, raft foundation [in m^3]

    steel_mass_short_ton
        (float) short tons of reinforcing steel

    material_needs_per_turbine
        (pd.DataFrame) table containing material needs info for -> Steel - rebar, Concrete 5000 psi, Excavated dirt, Backfill.

    operation_data
        (pd.DataFrame) TODO: What's the best one line definition for this?


    **TODO: Weather delay set of outputs -> ask Alicia for formal definitions of these keys.**

    total_foundation_cost
        (pd.DataFrame) summary of foundation costs (in USD) broken down into 4 main categories:
        1. Equipment Rental
        2. Labor
        3. Materials
        4. Mobilization
    """

    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """

        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name


        #Constants used in FoundationCost class. Making this data private (hidden from outside of this class): #TODO: Change private variables to protected.
        self._kg_per_tonne = 1000
        self._cubicm_per_cubicft = 0.0283168
        self._steel_density = 9490  # kg / m^3
        self._cubicyd_per_cubicm = 1.30795
        self._ton_per_tonne = 0.907185


    def calculate_foundation_load(self, foundation_load_input_data, foundation_load_output_data):
        """

        Function to calculate foundation load.

        Parameters
        -------
        Int Section height m

        Surface area sq (in m^2)

        Coeff drag (installed)

        Lever arm m (in m)

        Multplier drag rotor

        Multiplier tower drag

        Mass tonne



        Returns
        -------
        Dead load [in N] -> F_dead_kN_per_turbine

        Lateral load [in N] -> F_horiz_kN_per_turbine

        Moment [N.m] -> M_tot_kN_m_per_turbine

        Foundation radius based on overturning moment [in m] -> Radius_o_m

        Foundation radius based on slipping [in m] -> Radius_s_m

        Foundation radius based on gapping [in m] -> Radius_g_m

        Foundation radius based on bearing pressure [in m] -> Radius_b_m

        Largest foundation radius based on all three foundation design criteria (moment, gapping, bearing [in m]) -> Radius_m

        Raises
        ------
        ValueError
            Raises a value error if r_bearing is calculated to be a negative value.
        """
        # set exposure constants
        a = 9.5
        z_g = 274.32

        # get section height
        z = foundation_load_input_data['Section height m']

        # get cross-sectional area
        a_f = foundation_load_input_data['Surface area sq m']

        # get coefficient of drag
        c_d = foundation_load_input_data['Coeff drag (installed)']

        # get lever arm
        l = foundation_load_input_data['Lever arm m']

        # get multipliers for tower and rotor
        multiplier_rotor = foundation_load_input_data['Multplier drag rotor']
        multiplier_tower = foundation_load_input_data['Multiplier tower drag']


        # calculate wind pressure
        k_z = 2.01 * (z / z_g) ** (2 / a)  # exposure factor
        k_d = 0.95  # wind directionality factor
        k_zt = 1  # topographic factor
        v = foundation_load_input_data['gust_velocity_m_per_s']
        wind_pressure = 0.613 * k_z * k_zt * k_d * v ** 2

        # calculate wind loads on each tower component
        g = 0.85  # gust factor
        c_f = 0.6  # coefficient of force
        f_t = (wind_pressure * g * c_f * a_f) * multiplier_tower

        # calculate drag rotor
        rho = 1.225  # air density in kg/m^3
        f_r = (0.5 * rho * c_d * a_f * v ** 2) * multiplier_rotor

        f = (f_t + f_r)

        # calculate dead load in N
        g = 9.8  # m / s ^ 2
        f_dead = sum(foundation_load_input_data['Mass tonne']) * g * self._kg_per_tonne / 1.15  # scaling factor to adjust dead load for uplift

        # calculate moment from each component at base of tower
        m_overturn = f * l

        # get total lateral load (N) and moment (N * m)
        f_lat = f.sum()  # todo: add f_lat (drag force) to output csv
        m_overturn = m_overturn.sum()

        # compare to moment from rated thrust
        rated_thrust = foundation_load_input_data['rated_thrust_N']
        m_thrust = rated_thrust * max(l)
        m_tot = max(m_thrust, m_overturn)

        # compare lateral load to rated thrust
        f_horiz = max(f_lat, rated_thrust)

        # calculate foundation radius based on overturning moment
        vol_fraction_fill = 0.55
        vol_fraction_concrete = 1 - vol_fraction_fill
        safety_overturn = 1.5
        unit_weight_fill = 17.3e3  # in N / m^3
        unit_weight_concrete = 23.6e3  # in N / m^3
        bearing_pressure = foundation_load_input_data['bearing_pressure_n_m2']
        p = [(np.pi * foundation_load_input_data['depth'] * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete)), 0, f_dead, - (safety_overturn * (m_tot + f_horiz * foundation_load_input_data['depth']))]
        r_overturn = np.roots(p)
        r_overturn = np.real(r_overturn[np.isreal(r_overturn)])[0]

        # calculate foundation radius based on slipping
        safety_slipping = 1.5
        friction_angle_soil = 25
        tangent_slip_angle = math.tan((friction_angle_soil * math.pi) / 180)
        slipping_force_with_sf = (safety_slipping * f_lat)
        # first check if slipping is already satisfied by dead weight
        if slipping_force_with_sf < (f_dead * tangent_slip_angle):
            r_slipping = 0
        else:
            # Calculate foundation radius based on slipping:
            r_slipping = (((slipping_force_with_sf / tangent_slip_angle) - f_dead) /
                          ((vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) * math.pi * foundation_load_input_data['depth'])) ** 0.5

        r_test_gapping = max(r_overturn, r_slipping)

        # calculate foundation radius based on gapping
        # check if gapping constrain is already satisfied - r / 3 < e
        foundation_vol = np.pi * r_test_gapping ** 2 * foundation_load_input_data['depth']
        v_1 = (foundation_vol * (
                    vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) + f_dead)
        e = m_tot / v_1
        if (r_test_gapping / 3) < e:
            r_gapping = 0
        else:
            def r_g(x):
                foundation_vol = np.pi * x** 2 * foundation_load_input_data['depth']
                v_1 = (foundation_vol * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) + f_dead)
                e = m_tot / v_1
                return (e * 3 - x)
            result = root_scalar(r_g, method='brentq', bracket=[0.5*r_overturn, 50], xtol=1e-4, maxiter=50)
            r_gapping = result.root
            if not result.converged:
                raise ValueError(f'Warning {self.project_name} calculate_foundation_load r_gapping solve failed, {result.flag}')

        r_test_bearing = max(r_test_gapping, r_gapping)

        # calculate foundation radius based on bearing pressure
        def r_b(x):
            foundation_vol = np.pi * r_test_bearing ** 2 * foundation_load_input_data['depth']
            v_1 = (foundation_vol * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) + f_dead)
            e = m_tot / v_1
            a_eff = v_1 / bearing_pressure
            return (2 * (x ** 2 - e * (x ** 2 - e ** 2) ** 0.5) - a_eff)

        # Get minimum radius
        foundation_vol = np.pi * r_test_bearing ** 2 * foundation_load_input_data['depth']
        v_1 = (foundation_vol * (vol_fraction_fill * unit_weight_fill + vol_fraction_concrete * unit_weight_concrete) + f_dead)
        min_r = m_tot / v_1

        result = root_scalar(r_b, method='brentq', bracket=[min_r+1e-3, 50], xtol=1e-10, maxiter=50)
        r_bearing = result.root

        if not result.converged:
            raise ValueError(f'Warning {self.project_name} calculate_foundation_load r_bearing solve failed, {result.flag}')

        # pick the largest foundation radius based on all 4 foundation design criteria: moment, gapping, bearing, slipping
        r_choosen = max(r_bearing, r_overturn, r_slipping, r_gapping)

        foundation_load_output_data['F_dead_kN_per_turbine']    =   f_dead / 1e3
        foundation_load_output_data['F_horiz_kN_per_turbine']   =   f_lat / 1e3
        foundation_load_output_data['M_tot_kN_m_per_turbine']   =   m_tot / 1e3
        foundation_load_output_data['Radius_o_m']   =   r_overturn
        foundation_load_output_data['Radius_s_m']   =   r_slipping
        foundation_load_output_data['Radius_g_m']   =   r_gapping
        foundation_load_output_data['Radius_b_m']   =   r_bearing
        foundation_load_output_data['Radius_m']     =   r_choosen

        return foundation_load_output_data


    def determine_foundation_size(self, foundation_size_input_data, foundation_size_output_data):
        """
        Function to calculate the volume of a round, raft foundation. Assumes foundation made of concrete with 1 m thickness.

        Parameters
        -------
        Largest foundation radius based on all three foundation design criteria: moment, gapping, bearing [in m] -> Radius_m [in m]

        depth of foundation [in m] -> depth


        Returns
        -------
        Foundation volume [in m^3] -> foundation_volume_concrete_m3_per_turbine

        """
        #TODO: still updating/fine-tuning foundation size equations for small DW (Parangat - Feb 27, 2020)
        r = float(foundation_size_output_data['Radius_m'])
        if foundation_size_input_data['turbine_rating_MW'] < 0.1:
            foundation_size_output_data['excavated_volume_m3'] = r * r * foundation_size_input_data['depth']
            foundation_size_output_data['foundation_volume_concrete_m3_per_turbine'] = foundation_size_output_data['excavated_volume_m3'] * 0.45
        else:
            foundation_size_output_data['excavated_volume_m3'] = np.pi * (r + 0.5) ** 2 * foundation_size_input_data['depth']

            # only compute the portion of the foundation that is composed of concrete (45% concrete; other portion is
            # backfill); TODO: Add to sphinx -> (volume excavated = pi*(r_pick + .5m)^2 this assumes vertical sides which
            #  does not reflect reality as OSHA requires benched sides over 3’)
            foundation_size_output_data['foundation_volume_concrete_m3_per_turbine'] = np.pi * r ** 2 * \
                                                                                       foundation_size_input_data['depth'] * 0.45

        return foundation_size_output_data


    def estimate_material_needs_per_turbine(self, material_needs_per_turbine_input_data, material_needs_per_turbine_output_data):
        """
        Function to estimate amount of material based on foundation size and number of turbines.


        Parameters
        -------
        Foundation concrete volume [in m^3] -> foundation_volume_concrete_m3_per_turbine


        Returns
        -------

        (Returns pd.DataFrame) material_needs_per_turbine


        """

        steel_mass_short_ton_per_turbine = (material_needs_per_turbine_output_data['foundation_volume_concrete_m3_per_turbine'] * 0.012 * self._steel_density / self._kg_per_tonne)
        concrete_volume_cubic_yards_per_turbine = material_needs_per_turbine_output_data['foundation_volume_concrete_m3_per_turbine'] * 0.985 * self._cubicyd_per_cubicm

        #Assign values to output dictionary:
        material_needs_per_turbine_output_data['material_needs_per_turbine'] = pd.DataFrame([['Steel - rebar', steel_mass_short_ton_per_turbine, 'ton (short)'],
                                                                     ['Concrete 5000 psi', concrete_volume_cubic_yards_per_turbine, 'cubic yards'],
                                                                     ['Excavated dirt', material_needs_per_turbine_output_data['excavated_volume_m3'] * self._cubicyd_per_cubicm, 'cubic_yards'],
                                                                     ['Backfill', material_needs_per_turbine_output_data['excavated_volume_m3'] * self._cubicyd_per_cubicm, 'cubic_yards']],
                                                                    columns=['Material type ID', 'Quantity of material', 'Units'])

        material_needs_per_turbine_output_data['steel_mass_short_ton_per_turbine'] = steel_mass_short_ton_per_turbine
        return material_needs_per_turbine_output_data['material_needs_per_turbine']


    def estimate_construction_time(self, construction_time_input_data, construction_time_output_data):
        """
        Function to estimate construction time on per turbine basis. TODO: What's a better definition of this function. It's task is to return a pd.DataFrame (operation_data).

        Parameters
        -------
        duration_construction

        pd.DataFrame
            rsmeans

        pd.DataFrame
            material_needs_per_turbine



        Returns
        -------

        (pd.DataFrame) operation_data

        """

        foundation_construction_time = construction_time_input_data['construct_duration'] * 1 / 3
        #throughput_operations = construction_time_input_data['throughput_operations']
        throughput_operations = construction_time_input_data['rsmeans']
        material_needs_per_turbine = construction_time_output_data['material_needs_per_turbine']
        quantity_materials_entire_farm = material_needs_per_turbine['Quantity of material'] * construction_time_input_data['num_turbines']


        #Calculations for estimate construction time will be on entire wind farm basis:
        construction_time_output_data['material_needs_entire_farm'] = material_needs_per_turbine.copy()
        material_needs_entire_farm = construction_time_output_data['material_needs_entire_farm']
        material_needs_entire_farm['Quantity of material'] = quantity_materials_entire_farm
        if construction_time_input_data['turbine_rating_MW'] <= 0.1:
            operation_data = throughput_operations.where(
                throughput_operations['Module'] == 'Small DW Foundations').dropna(
                thresh=4)
        else:
            operation_data = throughput_operations.where(throughput_operations['Module'] == 'Foundations').dropna(
                thresh=4)

        #operation data for entire wind farm:
        operation_data = pd.merge(material_needs_entire_farm, operation_data, on=['Material type ID'], how='outer', sort=True)
        operation_data['Number of days'] = operation_data['Quantity of material'] / operation_data['Daily output']
        operation_data['Number of crews'] = np.ceil((operation_data['Number of days'] / 30) / foundation_construction_time)

        alpha = operation_data[operation_data['Type of cost'] == 'Labor']
        operation_data_id_days_crews_workers = alpha[['Operation ID', 'Number of days', 'Number of crews', 'Number of workers']]

        # if more than one crew needed to complete within construction duration then assume that all construction happens
        # within that window and use that timeframe for weather delays; if not, use the number of days calculated
        operation_data['time_construct_bool'] = operation_data['Number of days'] > foundation_construction_time * 30
        boolean_dictionary = {True: foundation_construction_time * 30, False: np.NAN}
        operation_data['time_construct_bool'] = operation_data['time_construct_bool'].map(boolean_dictionary)
        operation_data['Time construct days'] = operation_data[['time_construct_bool', 'Number of days']].min(axis=1)
        num_days = operation_data['Time construct days'].max()

        construction_time_output_data['operation_data_id_days_crews_workers'] = operation_data_id_days_crews_workers
        construction_time_output_data['operation_data_entire_farm'] = operation_data

        # pull out management data #TODO: Add this cost to Labor cost next
        if construction_time_input_data['turbine_rating_MW'] > 0.1:
            crew_cost = self.input_dict['crew_cost']
            crew = self.input_dict['crew'][self.input_dict['crew']['Crew type ID'].str.contains('M0')]
            management_crew = pd.merge(crew_cost, crew, on=['Labor type ID'], sort=True)
            management_crew = management_crew.assign(per_diem_total=management_crew['Per diem USD per day'] * management_crew['Number of workers'] * num_days)
            management_crew = management_crew.assign(hourly_costs_total=management_crew['Hourly rate USD per hour'] * self.input_dict['hour_day'][self.input_dict['time_construct']] * num_days)
            management_crew = management_crew.assign(total_crew_cost_before_wind_delay=management_crew['per_diem_total'] + management_crew['hourly_costs_total'])
            self.output_dict['management_crew'] = management_crew
            self.output_dict['managament_crew_cost_before_wind_delay'] = management_crew['total_crew_cost_before_wind_delay'].sum()
        else:
            self.output_dict['managament_crew_cost_before_wind_delay'] = 0

        return construction_time_output_data['operation_data_entire_farm']


    def calculate_weather_delay(self, weather_delay_input_data, weather_delay_output_data):
        """
        Function to calculate wind delay for foundations.

        Keys in weather_delay_input_data
        --------------------------------
        weather_window

        duration_construction

        start_delay

        critical_wind_speed

        operational_hrs_per_day

        height_interest

        wind_shear_exponent
        """

        # construct WeatherDelay module
        WD(weather_delay_input_data, weather_delay_output_data)

        # compute weather delay
        wind_delay = pd.DataFrame(weather_delay_output_data['wind_delays'])

        # if greater than 4 hour delay, then shut down for full day (10 hours)
        wind_delay[(wind_delay > 4)] = 10
        weather_delay_output_data['wind_delay_time'] = float(wind_delay.sum().iloc[0])

        return weather_delay_output_data


    def calculate_costs(self, calculate_costs_input_dict, calculate_costs_output_dict):
        """
        Function to calculate the total foundation cost.

        Keys in input dictionary
        ------------------------
        pd.DataFrame
            material_needs_per_turbine

        pd.DataFrame
            material_price

        pd.DataFrame
            operation_data

        wind_delay_time

        operational_hrs_per_day

        wind_multiplier

        pd.DataFrame
            rsmeans


        Returns
        -------

        (pd.DataFrame) total_foundation_cost


        """


        material_vol_entire_farm = calculate_costs_output_dict['material_needs_entire_farm']
        material_price = calculate_costs_input_dict['material_price']


        material_data_entire_farm = pd.merge(material_vol_entire_farm, material_price, on=['Material type ID'], sort=True)
        material_data_entire_farm['Cost USD'] = material_data_entire_farm['Quantity of material'] * pd.to_numeric(material_data_entire_farm['Material price USD per unit'])     # material data on a total wind farm basis


        operation_data = calculate_costs_output_dict['operation_data_entire_farm']

        wind_delay = calculate_costs_output_dict['wind_delay_time']

        wind_delay_fraction = (wind_delay / calculate_costs_input_dict['operational_hrs_per_day']) / operation_data['Time construct days'].max(skipna=True)
        # check if wind_delay_fraction is greater than 1, which would mean weather delays are longer than they can possibily be for the input data
        if wind_delay_fraction > 1:
            raise ValueError('{}: Error: Wind delay greater than 100%'.format(type(self).__name__))
        wind_multiplier = 1 / (1 - wind_delay_fraction)
        calculate_costs_output_dict['wind_multiplier'] = wind_multiplier

        rsmeans = calculate_costs_input_dict['rsmeans']
        if calculate_costs_input_dict['turbine_rating_MW'] > 0.1:
            rsmeans = rsmeans.where(rsmeans['Module'] == 'Foundations').dropna(thresh=4)
        else:
            rsmeans = rsmeans.where(rsmeans['Module'] == 'Small DW Foundations').dropna(thresh=4)

        labor_equip_data = pd.merge(material_vol_entire_farm, rsmeans, on=['Material type ID'], sort=True)

        # Create foundation cost dataframe
        foundation_cost = pd.DataFrame(columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # Calculate per diem
        per_diem = operation_data['Number of workers'] * operation_data['Number of crews'] * (operation_data['Time construct days'] +
                                                                                              np.ceil(operation_data['Time construct days'] / 7)) * calculate_costs_input_dict['rsmeans_per_diem']
        where_are_na_ns = np.isnan(per_diem)
        per_diem[where_are_na_ns] = 0
        labor_equip_data['Cost USD'] = (labor_equip_data['Quantity of material'] * labor_equip_data['Rate USD per unit'] * calculate_costs_input_dict['overtime_multiplier'] + per_diem + calculate_costs_output_dict['managament_crew_cost_before_wind_delay']) * wind_multiplier
        self.output_dict['labor_equip_data'] = labor_equip_data

        # EQUIPMENT COST
        # Create equipment costs row to be appended to foundation_cost
        equipment_dataframe = labor_equip_data[labor_equip_data['Type of cost'].str.match('Equipment rental')]

        equipment_cost_usd_without_delay = (
                equipment_dataframe['Quantity of material'] * equipment_dataframe['Rate USD per unit'] *
                calculate_costs_input_dict['overtime_multiplier'] + per_diem)
        equipment_cost_usd_with_weather_delays = equipment_cost_usd_without_delay.sum() * wind_multiplier
        equipment_costs = pd.DataFrame([['Equipment rental', equipment_cost_usd_with_weather_delays, 'Foundation']],
                                       columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # LABOR COST
        # Create labor costs row to be appended to foundation_cost
        labor_dataframe = labor_equip_data[labor_equip_data['Type of cost'].str.match('Labor')]
        labor_cost_usd_without_management= (labor_dataframe['Quantity of material'] * labor_dataframe['Rate USD per unit'] * calculate_costs_input_dict['overtime_multiplier'] + per_diem )
        labor_cost_usd_with_management = labor_cost_usd_without_management.sum() + calculate_costs_output_dict['managament_crew_cost_before_wind_delay']
        labor_cost_usd_with_management_plus_weather_delays = labor_cost_usd_with_management * wind_multiplier
        labor_costs = pd.DataFrame([['Labor', labor_cost_usd_with_management_plus_weather_delays, 'Foundation']],
                                   columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # MATERIAL COST
        material_cost_dataframe = pd.DataFrame(columns=['Operation ID', 'Type of cost', 'Cost USD'])
        material_cost_dataframe['Operation ID'] = material_data_entire_farm['Material type ID']
        material_cost_dataframe['Type of cost'] = 'Materials'
        material_cost_dataframe['Cost USD'] = material_data_entire_farm['Cost USD']
        material_costs_sum = material_cost_dataframe['Cost USD'].sum()
        material_costs = pd.DataFrame([['Materials', material_costs_sum, 'Foundation']],
                                      columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        # Append all cost items to foundation_cost
        foundation_cost = pd.concat( (foundation_cost,equipment_costs,labor_costs,material_costs) )

        # Calculate mobilization cost as percentage of total foundation cost and add to foundation_cost
        # Assumed 5% of total foundation cost and add to foundation_cost for utility scale plant
        # A function of turbine size for distributed wind (< 10 turbines)
        if calculate_costs_input_dict['num_turbines'] > 10:
            mobilization_cost = foundation_cost['Cost USD'].sum() * 0.05
        else:
            if calculate_costs_input_dict['turbine_rating_MW'] < 0.1:
                # Zero since mobilization cost of equipment is included in the equipment rental cost
                mobilization_cost = 0
            else:
                # There is mobilization cost for 0-10 turbines 100+ kW in rating.
                num_turbines = calculate_costs_input_dict['num_turbines']
                rating = calculate_costs_input_dict['turbine_rating_MW']
                mobilization_multipler = self.mobilization_cost_multiplier(rating)
                mobilization_cost = foundation_cost['Cost USD'].sum() / num_turbines * mobilization_multipler

        mob_cost = pd.DataFrame([['Mobilization', mobilization_cost, 'Foundation']], columns=['Type of cost', 'Cost USD', 'Phase of construction'])

        foundation_cost = pd.concat((foundation_cost, mob_cost))

        # todo: we add a separate tab in the output file for costs (all costs will be the same format but it's a different format than other data)
        # columns in cost tab would include project_id, module, operation_id, type_of_cost, total_or_per_turbine, cost_usd
        # an example row for this type of output would be "project1, FoundationCost, Rebar installation, Labor, total, 2.371127e+06"
        # total_foundation_cost = foundation_cost.groupby(by=['Type of cost']).sum().reset_index()
        # total_foundation_cost['Phase of construction'] = 'Foundations'
        # total_cost_summed_foundation = total_foundation_cost.sum(numeric_only=True)[0] # todo: add total_cost_summed_foundation to output dict

        total_foundation_cost = foundation_cost
        calculate_costs_output_dict['total_foundation_cost'] = total_foundation_cost

        self.output_dict['labor_equip_data'] = labor_equip_data

        return total_foundation_cost

    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """

        # Note that some values are cast with float() so that XlsxWriter
        # (the library depended on by XlsxGenerator) can output them as
        # numbers. XlsxWriter, interestingly, cannot handle np.float32()
        # types.

        result = []
        module = type(self).__name__
        result.append({
            'unit': '',
            'type': 'variable',
            'variable_df_key_col_name': 'wind_multiplier',
            'value': float(self.output_dict['wind_multiplier'])
        })
        result.append({
            'unit': 'kN',
            'type': 'variable',
            'variable_df_key_col_name': 'F_dead',
            'value': float(self.output_dict['F_dead_kN_per_turbine'])
        })
        result.append({
            'unit': 'kN',
            'type': 'variable',
            'variable_df_key_col_name': 'F_horiz',
            'value': float(self.output_dict['F_horiz_kN_per_turbine'])
        })
        result.append({
            'unit': 'kN_m',
            'type': 'variable',
            'variable_df_key_col_name': 'M_tot_kN',
            'value': float(self.output_dict['M_tot_kN_m_per_turbine'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius_o',
            'value': float(self.output_dict['Radius_o_m'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius_g',
            'value': float(self.output_dict['Radius_g_m'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius_b',
            'value': float(self.output_dict['Radius_b_m'])
        })
        result.append({
            'unit': 'm',
            'type': 'variable',
            'variable_df_key_col_name': 'Radius',
            'value': float(self.output_dict['Radius_m'])
        })
        result.append({
            'unit': 'short_ton',
            'type': 'variable',
            'variable_df_key_col_name': 'steel_mass_short_ton_per_turbine',
            'value': self.output_dict['steel_mass_short_ton_per_turbine']
        })
        # foundation_volume_concrete_m3_per_turbine
        result.append({
            'unit': 'm^3',
            'type': 'variable',
            'variable_df_key_col_name': 'foundation_volume_concrete_m3_per_turbine',
            'value': self.output_dict['foundation_volume_concrete_m3_per_turbine']
        })

        for row in self.output_dict['operation_data_id_days_crews_workers'].itertuples():
            dashed_row = '{}-{}-{}-{}'.format(row[1], math.ceil(row[2]), row[3], row[4])
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'operation_data: Operation ID-Number of days-Number of crews-Number of workers',
                'value': dashed_row
            })

        for row in self.output_dict['material_needs_per_turbine'].itertuples():
            # This must be formatted in Python
            dashed_row = '{}-{}-{:.2e}'.format(row[0], row[1], row[2])
            result.append({
                'unit': row[3],
                'type': 'dataframe',
                'variable_df_key_col_name': 'material_needs_per_turbine: {}'.format('-'.join(self.output_dict['material_needs_per_turbine'].columns[:-1])),
                'value': dashed_row
            })

        for row in self.output_dict['total_foundation_cost'].itertuples():
            dashed_row = '{} <--> {} <--> {}'.format(row[1], row[3], math.ceil(row[2]))
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'Type of Cost <--> Phase of Construction <--> Cost in USD ',
                'value': dashed_row,
                'last_number': row[2]
            })

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module

        self.output_dict['foundation_cost_csv'] = result
        return result

    def outputs_for_module_type_operation(self, input_dict, output_dict):
        result = []
        module = type(self).__name__

        costs_by_module_type_operation = self.output_dict['labor_equip_data']
        for _, row in costs_by_module_type_operation.iterrows():
            _dict = dict()
            row = row.to_dict()
            _dict['operation_id'] = row['Operation ID']
            _dict['type_of_cost'] = row['Type of cost']
            _dict['cost'] = row['Cost USD']
            result.append(_dict)

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module
            _dict['total_or_turbine'] = 'total'

        self.output_dict['foundation_module_type_operation'] = result
        return result

    def run_module(self):
        """
        Runs the FoundationCost module and populates the IO dictionaries with calculated values.

        Parameters
        ----------
        <None>

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            either returns a 0 if the module ran successfully, or it returns the error
            raised that caused the failure.

        """
        try:
            self.calculate_foundation_load(self.input_dict, self.output_dict)  # Returns foundation load
            self.determine_foundation_size(self.input_dict, self.output_dict)  # Returns foundation volume
            self.estimate_material_needs_per_turbine(self.input_dict, self.output_dict)  # Returns material volume
            operation_data = self.estimate_construction_time(self.input_dict, self.output_dict)  # Estimates construction time

            # pull only global inputs for weather delay from input_dict
            weather_data_keys = ('wind_shear_exponent',
                                 'weather_window')

            # specify foundation-specific weather delay inputs
            self.weather_input_dict = dict(
                [(i, self.input_dict[i]) for i in self.input_dict if i in set(weather_data_keys)])
            self.weather_input_dict[
                'start_delay_hours'] = 0  # assume zero start for when foundation construction begins (start at beginning of construction time)
            self.weather_input_dict[
                'critical_wind_speed_m_per_s'] = self.input_dict['critical_speed_non_erection_wind_delays_m_per_s']
            self.weather_input_dict[
                'wind_height_of_interest_m'] = self.input_dict['critical_height_non_erection_wind_delays_m']

            # duration_construction is in units of days
            # duration_construction_months is in units of months
            days_per_month = 30
            duration_construction = operation_data['Time construct days'].max(skipna=True)
            duration_construction_months = duration_construction / days_per_month
            self.output_dict['foundation_construction_months'] = duration_construction_months

            # compute and specify weather delay mission time for roads
            operational_hrs_per_day = self.input_dict['hour_day'][self.input_dict['time_construct']]
            mission_time_hrs = duration_construction * operational_hrs_per_day
            self.weather_input_dict['mission_time_hours'] = mission_time_hrs
            self.input_dict['operational_hrs_per_day'] = operational_hrs_per_day

            self.calculate_weather_delay(self.weather_input_dict, self.output_dict)
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            # self.output_dict['labor_equip_data']
            # self.output_dict['foundation_module_type_operation'] = self.outputs_for_module_type_operation(self.input_dict, self.output_dict)
            self.output_dict['foundation_module_type_operation'] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict['total_foundation_cost'],
                project_id=self.project_name,
                total_or_turbine=True
            )
            return 0, 0   # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} FoundationCost")
            return 1, error    # module did not run successfully



"""
A test where we can modify different parts of the modeling_options to ensure
they work with a top-to-bottom WISDEM run.
"""

import os
import unittest

from wisdem.glue_code.runWISDEM import run_wisdem

test_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
    + os.sep
    + "examples"
    + os.sep
    + "02_reference_turbines"
    + os.sep
)
fname_analysis_options = test_dir + "analysis_options.yaml"
this_dir = os.path.dirname(os.path.realpath(__file__))
fname_modeling_options = this_dir + os.sep + "modified_modeling_options.yaml"


class TestRegression(unittest.TestCase):
    def test15MW(self):
        ## IEA 15MW
        fname_wt_input = test_dir + "IEA-15-240-RWT.yaml"
        wt_opt, _, _ = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)

        self.assertAlmostEqual(wt_opt["rotorse.blade_mass"][0], 68206.4068005262, 1) # new value: improved interpolation
        self.assertAlmostEqual(wt_opt["rotorse.rp.AEP"][0] * 1.0e-6, 77.68139716114459, 1)
        self.assertAlmostEqual(wt_opt["financese.lcoe"][0] * 1.0e3, 86.5697873533143, 1)
        self.assertAlmostEqual(wt_opt["rotorse.rs.tip_pos.tip_deflection"][0], 25.2532523698, 1)
        self.assertAlmostEqual(wt_opt["towerse.z_param"][-1], 144.386, 3)


if __name__ == "__main__":
    unittest.main()



"""
A test where we can modify different parts of the modeling_options to ensure
they work with a top-to-bottom WISDEM run.
"""

import os
import unittest

from wisdem.glue_code.runWISDEM import run_wisdem

test_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
    + os.sep
    + "examples"
    + os.sep
    + "09_floating"
    + os.sep
)
fname_analysis_options = test_dir + "analysis_options.yaml"
fname_modeling_options = test_dir + "modeling_options.yaml"


class TestRegression(unittest.TestCase):
    def test15MW(self):
        ## IEA 15MW
        fname_wt_input = test_dir + "IEA-15-240-RWT_VolturnUS-S.yaml"
        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, fname_analysis_options
        )

        self.assertAlmostEqual(wt_opt["rotorse.rp.AEP"][0] * 1.0e-6, 77.8888558791, 1)
        self.assertAlmostEqual(wt_opt["rotorse.blade_mass"][0], 68206.4068005262, 1) # new value: improved interpolation
        self.assertAlmostEqual(wt_opt["financese.lcoe"][0] * 1.0e3, 94.29792690850104, 1)
        self.assertAlmostEqual(wt_opt["rotorse.rs.tip_pos.tip_deflection"][0], 25.8310690720, 1)


if __name__ == "__main__":
    unittest.main()
